@@ -3,6 +3,7 @@
 #include "Asm.h"
 #include "../OpTable.h"
 #include "../Exception.h"
+#include "Utils/Cache.h"
 
 namespace code {
 	namespace arm64 {
@@ -237,6 +238,12 @@ namespace code {
 		}
 
 		void prologOut(Output *to, Instr *instr) {
+#if defined(ARM_USE_PAC)
+			// PACIASP - to "sign" the pointer using current value of SP.
+			to->putInt(0xD503233F);
+			to->markReturnAuth();
+#endif
+
 			Offset stackSize = instr->src().offset();
 			Int scaled = stackSize.v64() / 8;
 			if (isImm7S(scaled)) {
@@ -299,6 +306,11 @@ namespace code {
 				throw new (to) InvalidValue(S("Too large stack size for Arm64!"));
 			}
 
+#if defined(ARM_USE_PAC)
+			// AUTIASP
+			to->putInt(0xD50323BF);
+#endif
+
 			// Note: No DWARF metadata since this could be an early return.
 		}
 
@@ -309,6 +321,8 @@ namespace code {
 			Reg dest1 = instr->dest().reg();
 			Reg dest2 = noReg;
 
+			Bool isDynamic = instr->src().hasOffsetRef();
+
 			Bool intReg = isIntReg(dest1);
 
 			// Bytes are special:
@@ -317,28 +331,33 @@ namespace code {
 					putLoadStoreMidS(to, intReg ? 0x1C2 : 0x1E2, intRegSP(baseReg), intRegZR(dest1), offset);
 				else
 					putLoadStoreLargeU(to, intReg ? 0x0E5 : 0x0F5, intRegSP(baseReg), intRegZR(dest1), offset);
+				if (isDynamic)
+					to->markOffset(ArmOffsetUpdater::tLoadStore, instr->src().offsetRef());
 				return false;
 			}
 
 			// Look at "next" to see if we can merge it with this instruction.
 			if (next && next->dest().type() == opRegister && next->src().type() == opRelative) {
 				if (same(next->src().reg(), baseReg) && Int(next->dest().size().size64()) == opSize) {
-					// Note: It is undefined to load the same register multiple times. Also: it
-					// might break semantics when turning:
-					// - ldr x0, [x0]
-					// - ldr x0, [x0+8]
-					// into
-					// - ldp x0, x0, [x0]
-					if (!same(next->dest().reg(), dest1) && isIntReg(next->dest().reg()) == intReg) {
-						// Look at the offsets, if they are next to each other, we can merge them.
-						Int off = next->src().offset().v64();
-						if (off == offset + opSize && isImm7S(offset / opSize)) {
-							dest2 = next->dest().reg();
-						} else if (off == offset - opSize && isImm7S(off / opSize)) {
-							// Put the second one first.
-							dest2 = dest1;
-							dest1 = next->dest().reg();
-							offset = off;
+					// We don't support offset references. They would invalidate the tight size we have.
+					if (!isDynamic && !next->src().hasOffsetRef()) {
+						// Note: It is undefined to load the same register multiple times. Also: it
+						// might break semantics when turning:
+						// - ldr x0, [x0]
+						// - ldr x0, [x0+8]
+						// into
+						// - ldp x0, x0, [x0]
+						if (!same(next->dest().reg(), dest1) && isIntReg(next->dest().reg()) == intReg) {
+							// Look at the offsets, if they are next to each other, we can merge them.
+							Int off = next->src().offset().v64();
+							if (off == offset + opSize && isImm7S(offset / opSize)) {
+								dest2 = next->dest().reg();
+							} else if (off == offset - opSize && isImm7S(off / opSize)) {
+								// Put the second one first.
+								dest2 = dest1;
+								dest1 = next->dest().reg();
+								offset = off;
+							}
 						}
 					}
 				}
@@ -368,6 +387,9 @@ namespace code {
 					putLoadStoreLargeU(to, opSize == 4 ? 0x2F5 : 0x3F5, intRegSP(baseReg), fpReg(dest1), offset);
 			}
 
+			if (isDynamic)
+				to->markOffset(ArmOffsetUpdater::tLoadStore, instr->src().offsetRef());
+
 			return dest2 != noReg;
 		}
 
@@ -378,6 +400,8 @@ namespace code {
 			Reg src1 = instr->src().reg();
 			Reg src2 = noReg;
 
+			Bool isDynamic = instr->dest().hasOffsetRef();
+
 			Bool intReg = isIntReg(src1);
 
 			// Bytes are special:
@@ -386,24 +410,29 @@ namespace code {
 					putLoadStoreMidS(to, intReg ? 0x1C0 : 0x1E0, intRegSP(baseReg), intRegZR(src1), offset);
 				else
 					putLoadStoreLargeU(to, intReg ? 0x0E4 : 0x0F4, intRegSP(baseReg), intRegZR(src1), offset);
+				if (isDynamic)
+					to->markOffset(ArmOffsetUpdater::tLoadStore, instr->dest().offsetRef());
 				return false;
 			}
 
 			// Look at "next" to see if we can merge it with this instruction.
 			if (next && next->src().type() == opRegister && next->dest().type() == opRelative) {
 				if (same(next->dest().reg(), baseReg) && Int(next->src().size().size64()) == opSize) {
-					// Note: Contrary to the load instruction, it is well-defined to store the same
-					// register multiple times.
-					if (isIntReg(next->src().reg()) == intReg) {
-						// Look at the offsets, if they are next to each other, we can merge them.
-						Int off = next->dest().offset().v64();
-						if (off == offset + opSize && isImm7S(offset / opSize)) {
-							src2 = next->src().reg();
-						} else if (off == offset - opSize && isImm7S(off / opSize)) {
-							// Put the second one first.
-							src2 = src1;
-							src1 = next->src().reg();
-							offset = off;
+					// We don't support offset references. They would invalidate the tight size we have.
+					if (!isDynamic && !next->dest().hasOffsetRef()) {
+						// Note: Contrary to the load instruction, it is well-defined to store the same
+						// register multiple times.
+						if (isIntReg(next->src().reg()) == intReg) {
+							// Look at the offsets, if they are next to each other, we can merge them.
+							Int off = next->dest().offset().v64();
+							if (off == offset + opSize && isImm7S(offset / opSize)) {
+								src2 = next->src().reg();
+							} else if (off == offset - opSize && isImm7S(off / opSize)) {
+								// Put the second one first.
+								src2 = src1;
+								src1 = next->src().reg();
+								offset = off;
+							}
 						}
 					}
 				}
@@ -433,6 +462,9 @@ namespace code {
 					putLoadStoreLargeU(to, opSize == 4 ? 0x2F4 : 0x3F4, intRegSP(baseReg), fpReg(src1), offset);
 			}
 
+			if (isDynamic)
+				to->markOffset(ArmOffsetUpdater::tLoadStore, instr->dest().offsetRef());
+
 			return src2 != noReg;
 		}
 
@@ -448,9 +480,9 @@ namespace code {
 					putData3(to, 0x150, intRegZR(dest), 31, intRegZR(src), 0);
 			} else if (!intSrc && !intDst) {
 				if (size(src).size64() > 4)
-					to->putInt(0x1E602000 | (fpReg(src) << 5) | fpReg(dest));
+					to->putInt(0x1E604000 | (fpReg(src) << 5) | fpReg(dest));
 				else
-					to->putInt(0x1E202000 | (fpReg(src) << 5) | fpReg(dest));
+					to->putInt(0x1E204000 | (fpReg(src) << 5) | fpReg(dest));
 			} else if (intSrc) {
 				if (size(src).size64() > 4)
 					to->putInt(0x9E670000 | (intRegZR(src) << 5) | fpReg(dest));
@@ -464,6 +496,19 @@ namespace code {
 			}
 		}
 
+		// Helper to load an immediate value to a register.
+		static void loadImm(Output *to, Word imm, Reg dest) {
+			if (imm <= 0xFFFF) {
+				// Small enough to use MOVZ
+				to->putInt(0xD2800000 | ((0xFFFF & imm) << 5) | intRegZR(dest));
+			} else if (~imm <= 0xFFFF) {
+				// Store inverse using MOVN
+				to->putInt(0x92800000 | ((0xFFFF & ~imm) << 5) | intRegZR(dest));
+			} else {
+				throw new (to) InvalidValue(TO_S(to, S("Too large immediate to load: ") << imm));
+			}
+		}
+
 		// Special version called directly when more than one mov was found. Returns "true" if we
 		// could merge the two passed to us. We know that "next" is a mov op if it is non-null.
 		bool movOut(Output *to, Instr *instr, MAYBE(Instr *) next) {
@@ -498,10 +543,12 @@ namespace code {
 				loadLongConst(to, intRegZR(destReg), src.object());
 				return false;
 			case opConstant:
-				if (src.constant() > 0xFFFF)
-					throw new (to) InvalidValue(TO_S(to, S("Too large immediate to load: ") << src));
 				// Note: No difference between nat and word version.
-				to->putInt(0xD2800000 | ((0xFFFF & src.constant()) << 5) | intRegZR(destReg));
+				loadImm(to, src.constant(), destReg);
+				return false;
+			case opOffReference:
+				loadImm(to, src.offsetRef().offset().v64(), destReg);
+				to->markOffset(ArmOffsetUpdater::tImm, src.offsetRef());
 				return false;
 			case opLabel:
 			case opRelativeLbl: {
@@ -523,6 +570,10 @@ namespace code {
 			movOut(to, instr, null);
 		}
 
+		void shadowMovOut(Output *to, Instr *instr) {
+			movOut(to, instr);
+		}
+
 		void leaOut(Output *to, Instr *instr) {
 			Operand destOp = instr->dest();
 			if (destOp.type() != opRegister)
@@ -541,6 +592,8 @@ namespace code {
 					// sub:
 					putData2(to, 0x344, dest, intRegZR(src.reg()), -src.offset().v64());
 				}
+				if (src.hasOffsetRef())
+					to->markOffset(ArmOffsetUpdater::tLea, src.offsetRef());
 				break;
 			case opReference:
 				// This means to load the refSource instead of loading from the pointer.
@@ -554,7 +607,7 @@ namespace code {
 		void callOut(Output *to, Instr *instr) {
 			// Note: We need to use x17 for temporary values. This is assumed by the code in Gc/CodeArm64.cpp.
 
-			Nat offset;
+			Int offset;
 			Operand target = instr->src();
 			switch (target.type()) {
 			case opReference:
@@ -570,8 +623,16 @@ namespace code {
 				break;
 			case opRelative:
 				// Split into two op-codes: a load and a register call.
-				offset = target.offset().v64() / 8;
-				putLoadStoreLargeU(to, 0x3E5, intRegZR(target.reg()), 17, offset);
+				offset = target.offset().v64();
+				if (offset >= 0) {
+					Nat uOffset = Nat(offset) / 8;
+					putLoadStoreLargeU(to, 0x3E5, intRegSP(target.reg()), 17, uOffset);
+				} else {
+					putLoadStoreMidS(to, 0x7C2, intRegSP(target.reg()), 17, offset);
+				}
+				if (target.hasOffsetRef())
+					to->markOffset(ArmOffsetUpdater::tLoadStore, target.offsetRef());
+
 				// blr x17
 				to->putInt(0xD63F0220);
 				break;
@@ -631,13 +692,20 @@ namespace code {
 				to->putInt(0xD61F0000 | (intRegZR(target.reg()) << 5));
 				break;
 			case opRelative:
-				// Split into two op-codes: a load and a register jump.
-				offset = target.offset().v64() / 8;
-				putLoadStoreLargeU(to, 0x3E5, intRegZR(target.reg()), 17, offset);
+				offset = target.offset().v64();
+				if (offset >= 0) {
+					Nat uOffset = Nat(offset) / 8;
+					putLoadStoreLargeU(to, 0x3E5, intRegSP(target.reg()), 17, uOffset);
+				} else {
+					putLoadStoreMidS(to, 0x7C2, intRegSP(target.reg()), 17, offset);
+				}
+				if (target.hasOffsetRef())
+					to->markOffset(ArmOffsetUpdater::tLoadStore, target.offsetRef());
 				// br x17
 				to->putInt(0xD61F0220);
 				break;
 			case opLabel:
+				to->markJump(target.label());
 				offset = Int(to->offset(target.label())) - Int(to->tell());
 				offset /= 4;
 				checkImm26S(to, offset);
@@ -651,23 +719,31 @@ namespace code {
 
 		// Generic output of data instructions that use 12-bit immediates or registers. Assumes that
 		// the high bit of the op-code is the size bit.
-		static void data12Out(Output *to, Instr *instr, Nat opImm, Nat opReg) {
+		static void data12Out(Output *to, Instr *instr, Nat opImm, Nat opReg, Nat opSpReg) {
 			assert(instr->dest().type() == opRegister,
 				L"Destinations for data operations should have been transformed into registers.");
-			Nat dest = intRegSP(instr->dest().reg());
-
 			if (instr->src().size().size64() >= 8) {
 				opImm |= 0x200;
 				opReg |= 0x400;
+				opSpReg |= 0x400;
 			}
 
+			Reg destReg = instr->dest().reg();
 			switch (instr->src().type()) {
 			case opRegister:
-				putData3(to, opReg, dest, dest, intRegSP(instr->src().reg()), 0);
+				if (same(destReg, ptrStack)) {
+					// Note "imm" is really option + imm3 here.
+					putData3(to, opSpReg, 31, 31, intRegZR(instr->src().reg()), 0x18);
+				} else {
+					Nat dest = intRegZR(destReg);
+					putData3(to, opReg, dest, dest, intRegZR(instr->src().reg()), 0);
+				}
 				break;
-			case opConstant:
+			case opConstant: {
+				Nat dest = intRegSP(destReg);
 				putData2(to, opImm, dest, dest, instr->src().constant());
 				break;
+			}
 			default:
 				assert(false, L"Unsupported source for data operation.");
 				break;
@@ -675,11 +751,11 @@ namespace code {
 		}
 
 		void addOut(Output *to, Instr *instr) {
-			data12Out(to, instr, 0x044, 0x058);
+			data12Out(to, instr, 0x044, 0x058, 0x059);
 		}
 
 		void subOut(Output *to, Instr *instr) {
-			data12Out(to, instr, 0x144, 0x258);
+			data12Out(to, instr, 0x144, 0x258, 0x259);
 		}
 
 		void cmpOut(Output *to, Instr *instr) {
@@ -1074,7 +1150,7 @@ namespace code {
 		}
 
 		void fcastOut(Output *to, Instr *instr) {
-			Bool in64 = instr->dest().size().size64() > 4;
+			Bool in64 = instr->src().size().size64() > 4;
 			Bool out64 = instr->dest().size().size64() > 4;
 
 			if (in64 == out64) {
@@ -1088,7 +1164,7 @@ namespace code {
 				op |= Nat(1) << 22;
 			if (out64)
 				op |= Nat(1) << 15;
-			op |= intRegZR(instr->dest().reg());
+			op |= fpReg(instr->dest().reg());
 			op |= fpReg(instr->src().reg()) << 5;
 			to->putInt(op);
 		}
@@ -1182,6 +1258,8 @@ namespace code {
 			OUTPUT(prolog),
 			OUTPUT(epilog),
 			OUTPUT(mov),
+			// Note: This is rare, so we don't bother with considering that it is the same as 'mov'.
+			OUTPUT(shadowMov),
 			OUTPUT(lea),
 			OUTPUT(call),
 			OUTPUT(ret),
@@ -1302,5 +1380,80 @@ namespace code {
 			to->mark(src->labels(src->count()));
 		}
 
+
+		/**
+		 * Updater.
+		 */
+
+		ArmOffsetUpdater::ArmOffsetUpdater(OffsetRef src, Content *inside, void *code, Nat codeOffset, Nat type)
+			: OffsetReference(src, inside), code(code), codeOffset(codeOffset), type(type) {
+			moved(offset());
+		}
+
+		void ArmOffsetUpdater::moved(Offset offset) {
+			Int value = offset.v64();
+			Nat uValue = Nat(value);
+
+			Nat *addr = (Nat *)((char *)code + codeOffset);
+			Nat oldOp = *addr;
+
+
+			switch (type) {
+			case tImm: {
+				Nat destReg = oldOp & 0x1F;
+				// Instr should be MOVZ or MOVN:
+				if (uValue <= 0xFFFF) {
+					*addr = 0xD2800000 | ((0xFFFF & uValue) << 5) | destReg;
+				} else if (~uValue <= 0xFFFF) {
+					*addr = 0x92800000 | ((0xFFFF & ~uValue) << 5) | destReg;
+				}
+				break;
+			}
+			case tLea: {
+				// Either an add or a sub instruction, depending on the sign of the value:
+				Nat opBits = oldOp & 0xFFC003FF;
+				if (value < 0) {
+					// Make sure we have a SUB instruction:
+					opBits |= Nat(0x40000000);
+					uValue = Nat(-value);
+				} else {
+					opBits &= ~Nat(0x40000000);
+				}
+
+				*addr = opBits | ((uValue & 0xFFF) << 10);
+
+				break;
+			}
+			case tLoadStore: {
+				// We keep the top 12 bits (op-code) and bottom 10 bits (src and dest regs).
+				// Note that LDUR have a 13 bit op-code, but the last bit is zero. Apart from
+				// that, the only difference is bit 24, which is set for LDUR and clear for LDR.
+				Nat opBits = oldOp & 0xFFC003FF;
+				Nat size = (oldOp >> 30) & 0x3;
+
+
+				if (value < 0) {
+					// LDUR
+					opBits &= ~Nat(0x01000000);
+					*addr = opBits | ((uValue & 0x1FF) << 12);
+				} else {
+					// LDR
+					opBits |= Nat(0x01000000);
+					uValue >>= size; // Scale the size
+					*addr = opBits | ((uValue & 0xFFF) << 10);
+				}
+				break;
+			}
+			default:
+				assert(false, L"Unknown ARM offset type.");
+				break;
+			}
+
+			// Note: This is technically not enough - other threads/cores need to run some
+			// instructions as well. This will typically be the case whenever it is safe to update
+			// offsets to data structures anyway.
+			invalidateSingleICache(addr);
+		}
+
 	}
 }
