@@ -26,6 +26,9 @@ namespace code {
 		// Output.
 		void STORM_FN toS(StrBuf *to) const;
 
+		// Create from an int. Mostly for backends, and low-level code.
+		static Block fromNat(Nat v) { return Block(v); }
+
 	private:
 		friend class Listing;
 		friend class Operand;
