@@ -216,6 +216,10 @@ namespace code {
 		// Get parameter info about a variable.
 		MAYBE(TypeDesc *) STORM_FN paramDesc(Var v) const;
 
+		// Get the parameter index of a variable. Performs a linear search, so might not be
+		// performant for large parameter lists.
+		Nat STORM_FN paramIndex(Var v) const;
+
 		// Get debug info about a variable.
 		MAYBE(VarInfo *) STORM_FN varInfo(Var v) const;
 
