@@ -29,5 +29,8 @@ namespace code {
 		// Convert to DWARF registers.
 		Nat dwarfRegister(Reg reg);
 
+		// Convert from DWARF registers.
+		Reg fromDwarfRegister(Nat id);
+
 	}
 }
