@@ -16,6 +16,10 @@ namespace storm {
 			add(expr);
 		}
 
+		Actuals::Actuals(Array<Expr *> *exprs) {
+			expressions = new (this) Array<Expr *>(*exprs);
+		}
+
 		Actuals::Actuals(const Actuals &o) : ObjectOn<Compiler>(o) {
 			expressions = new (this) Array<Expr *>(*o.expressions);
 		}
