@@ -27,6 +27,7 @@
 #include "Exception.h"
 #include "SystemException.h"
 #include "OS/StackTrace.h"
+#include "BuiltInSource.h"
 
 // Only included from here:
 #include "OS/SharedMaster.h"
@@ -97,8 +98,11 @@ namespace storm {
 		static const TypeFlags v = TypeFlags(typeValue | pod | simple);
 	};
 
+	static Url *defaultCreateRoot(Engine &e, void *root) {
+		return parsePathAsDir(e, (const wchar *)root);
+	}
 
-	Engine::Engine(const Path &root, ThreadMode mode, void *stackBase) :
+	Engine::Engine(const wchar *root, ThreadMode mode, void *stackBase) :
 		id(atomicIncrement(engineId)),
 		gc(defaultArena, defaultFinalizer),
 		threadGroup(util::memberVoidFn(this, &Engine::attachThread), util::memberVoidFn(this, &Engine::detachThread)),
@@ -107,6 +111,22 @@ namespace storm {
 		ioThread(null),
 		stackInfo(gc) {
 
+		create(&defaultCreateRoot, (void *)root, mode, stackBase);
+	}
+
+	Engine::Engine(CreateRootUrl callback, void *param, ThreadMode mode, void *stackBase) :
+		id(atomicIncrement(engineId)),
+		gc(defaultArena, defaultFinalizer),
+		threadGroup(util::memberVoidFn(this, &Engine::attachThread), util::memberVoidFn(this, &Engine::detachThread)),
+		world(gc),
+		objRoot(null),
+		ioThread(null),
+		stackInfo(gc) {
+
+		create(callback, param, mode, stackBase);
+	}
+
+	void Engine::create(CreateRootUrl createRoot, void *createParam, ThreadMode mode, void *stackBase) {
 		// Handle system errors.
 		setupSystemExceptions();
 
@@ -141,11 +161,8 @@ namespace storm {
 			Thread::stormType(*this)->setType(world.threads[0]);
 
 			// Set proper paths for the now created packages.
-#ifdef WINDOWS
-			o.root->setUrl(parsePath(*this, toS(root).c_str())->updated());
-#else
-			o.root->setUrl(parsePath(*this, toWChar(*this, toS(root).c_str())->v)->updated());
-#endif
+			Url *rootUrl = (*createRoot)(*this, createParam);
+			o.root->setUrl(rootUrl->updated());
 
 			// Done booting.
 			advance(bootDone);
@@ -389,75 +406,63 @@ namespace storm {
 		new (Place(obj)) Variant(src);
 	}
 
+	static const void *CODECALL refSourceAddr(code::RefSource *src) {
+		return src->address();
+	}
+
+	static Object *CODECALL cloneEnvGet(CloneEnv *env, Object *key) {
+		return env->cloned(key);
+	}
+
+	static void CODECALL cloneEnvPut(CloneEnv *env, Object *key, Object *val) {
+		env->cloned(key, val);
+	}
+
 	code::RefSource *Engine::createRef(builtin::BuiltIn ref) {
-#define W(x) S(x)
-#define FNREF(x) arena()->externalSource(S("C++:") W(#x), address(&x))
+#define FNREF(id, fn) case builtin::id:									\
+			return new (*this) BuiltInSource(builtin::id, address(&fn))
 
 		switch (ref) {
+			FNREF(lazyCodeUpdate, LazyCode::updateCode);
+			FNREF(ruleThrow, syntax::Node::throwError);
+			FNREF(alloc, allocType);
+			FNREF(allocArray, allocArray);
+			FNREF(as, stormAs);
+			FNREF(mapAtValue, MapBase::atRawValue);
+			FNREF(mapAtClass, MapBase::atRawClass);
+			FNREF(mapGetUnsafe, MapBase::getUnsafeRaw);
+			FNREF(setGetUnsafe, SetBase::getUnsafeRaw);
+			FNREF(enumToS, Enum::toString);
+			FNREF(futurePost, FutureBase::postRaw);
+			FNREF(futureResult, FutureBase::resultRaw);
+			FNREF(futureNoClone, FutureBase::setNoClone);
+			FNREF(spawnResult, spawnThreadResult);
+			FNREF(spawnFuture, spawnThreadFuture);
+			FNREF(spawnFutureId, spawnThreadId);
+			FNREF(spawnVoid, spawnThreadVoid);
+			FNREF(spawnVoidId, spawnThreadVoidId);
+			FNREF(fnNeedsCopy, FnBase::needsCopy);
+			FNREF(fnCall, fnCallRaw);
+			FNREF(fnCreate, fnCreateRaw);
+			FNREF(fnNull, fnNull);
+			FNREF(maybeToS, MaybeValueType::toSHelper);
+			FNREF(globalAddr, GlobalVar::dataPtr);
+			FNREF(throwAbstractError, throwAbstractError);
+			FNREF(throwException, throwException);
+			FNREF(createValVariant, createValVariant);
+			FNREF(createClassVariant, createClassVariant);
+			FNREF(createJoin, createJoin);
+			FNREF(refSourceAddr, refSourceAddr);
+			FNREF(cloneEnvGet, cloneEnvGet);
+			FNREF(cloneEnvPut, cloneEnvPut);
 		case builtin::engine:
 			return arena()->externalSource(S("engine"), this);
-		case builtin::lazyCodeUpdate:
-			return FNREF(LazyCode::updateCode);
-		case builtin::ruleThrow:
-			return FNREF(syntax::Node::throwError);
-		case builtin::alloc:
-			return FNREF(allocType);
-		case builtin::allocArray:
-			return FNREF(allocArray);
-		case builtin::as:
-			return FNREF(stormAs);
 		case builtin::VTableAllocOffset:
+			// TODO: An offsetRef?
 			return arena()->externalSource(S("vtableAllocOffset"), (const void *)VTableCpp::vtableAllocOffset());
 		case builtin::TObjectOffset:
+			// TODO: An offsetRef?
 			return arena()->externalSource(S("threadOffset"), (const void *)OFFSET_OF(TObject, thread));
-		case builtin::mapAtValue:
-			return FNREF(MapBase::atRawValue);
-		case builtin::mapAtClass:
-			return FNREF(MapBase::atRawClass);
-		case builtin::mapGetUnsafe:
-			return FNREF(MapBase::getUnsafeRaw);
-		case builtin::setGetUnsafe:
-			return FNREF(SetBase::getUnsafeRaw);
-		case builtin::enumToS:
-			return FNREF(Enum::toString);
-		case builtin::futurePost:
-			return FNREF(FutureBase::postRaw);
-		case builtin::futureResult:
-			return FNREF(FutureBase::resultRaw);
-		case builtin::futureNoClone:
-			return FNREF(FutureBase::setNoClone);
-		case builtin::spawnResult:
-			return FNREF(spawnThreadResult);
-		case builtin::spawnFuture:
-			return FNREF(spawnThreadFuture);
-		case builtin::spawnFutureId:
-			return FNREF(spawnThreadId);
-		case builtin::spawnVoid:
-			return FNREF(spawnThreadVoid);
-		case builtin::spawnVoidId:
-			return FNREF(spawnThreadVoidId);
-		case builtin::fnNeedsCopy:
-			return FNREF(FnBase::needsCopy);
-		case builtin::fnCall:
-			return FNREF(fnCallRaw);
-		case builtin::fnCreate:
-			return FNREF(fnCreateRaw);
-		case builtin::fnNull:
-			return FNREF(fnNull);
-		case builtin::maybeToS:
-			return FNREF(MaybeValueType::toSHelper);
-		case builtin::globalAddr:
-			return FNREF(GlobalVar::dataPtr);
-		case builtin::throwAbstractError:
-			return FNREF(throwAbstractError);
-		case builtin::throwException:
-			return FNREF(throwException);
-		case builtin::createValVariant:
-			return FNREF(createValVariant);
-		case builtin::createClassVariant:
-			return FNREF(createClassVariant);
-		case builtin::createJoin:
-			return FNREF(createJoin);
 		default:
 			assert(false, L"Unknown reference: " + ::toS(ref));
 			return null;
@@ -558,7 +563,7 @@ namespace storm {
 		if (!o.readObj) {
 			Value t(Object::stormType(*this));
 			o.readObj = nativeFunction(*this, t, S("_read_"), valList(*this, 1, t.asRef()), address(&readObject));
-			o.readObj->parentLookup = o.root;
+			o.readObj->parentLookup(o.root);
 		}
 		return o.readObj;
 	}
@@ -567,7 +572,7 @@ namespace storm {
 		if (!o.readTObj) {
 			Value t(TObject::stormType(*this));
 			o.readTObj = nativeFunction(*this, t, S("_readT_"), valList(*this, 1, t.asRef()), address(&readObject));
-			o.readTObj->parentLookup = o.root;
+			o.readTObj->parentLookup(o.root);
 		}
 		return o.readTObj;
 	}
