@@ -7,6 +7,8 @@
 #include "Core/Str.h"
 #include "Lib/Future.h"
 #include "Lib/Clone.h"
+#include "ReplaceActive.h"
+#include "Compiler/Lib/Io.h"
 
 namespace storm {
 
@@ -16,6 +18,7 @@ namespace storm {
 		lookupRef(null),
 		codeRef(null),
 		runOnThread(null),
+		toReplace(null),
 		myFlags(fnNone) {}
 
 	Function::Function(SrcPos pos, Value result, Str *name, Array<Value> *params) :
@@ -24,6 +27,7 @@ namespace storm {
 		lookupRef(null),
 		codeRef(null),
 		runOnThread(null),
+		toReplace(null),
 		myFlags(fnNone) {
 
 		this->pos = pos;
@@ -112,6 +116,123 @@ namespace storm {
 		}
 	}
 
+	MAYBE(Str *) Function::canReplace(Named *old, ReplaceContext *ctx) {
+		Function *oldFn = as<Function>(old);
+		if (!oldFn)
+			return new (this) Str(S("Unable to replace something that is not a function with a function."));
+
+		if (params->count() != oldFn->params->count())
+			return new (this) Str(S("Cannot modify the number of formal parameters to a function."));
+		for (Nat i = 0; i < params->count(); i++)
+			if (!ctx->normalize(params->at(i)).mayStore(ctx->normalize(oldFn->params->at(i))))
+				return new (this) Str(S("Cannot modify the types of formal parameters to incompatible types."));
+		if (!ctx->normalize(oldFn->result).mayStore(ctx->normalize(result)))
+			return new (this) Str(S("Cannot modify the return type of a function to an incompatible type."));
+
+		return null;
+	}
+
+	static void addOldFunction(Engine &e, GcWeakArray<const void> *&array, const void *ptr) {
+		size_t newCount = 0;
+		if (array) {
+			// Note: Safe to not use atomics here - this is an upper bound anyway.
+			for (size_t i = 0; i < array->count(); i++)
+				if (array->v[i])
+					newCount++;
+		}
+
+		GcWeakArray<const void> *n = runtime::allocWeakArray<const void>(e, newCount + 1);
+
+		size_t target = 0;
+		if (array) {
+			for (size_t i = 0; i < array->count(); i++) {
+				if (const void *p = array->read(i))
+					n->write(target++, p);
+			}
+		}
+
+		n->write(target++, ptr);
+		array = n;
+	}
+
+	void Function::doReplace(Named *old, ReplaceTasks *tasks, ReplaceContext *ctx) {
+		Function *f = (Function *)old;
+
+		// Store the old pointer to the function so that we can replace it if necessary.
+		// Note: We ignore 'lookup' completely - we don't expect much to happen there.
+		const void *oldCode = null;
+		if (f->codeRef) {
+			// Only consider cases where the code is GeneratedCode. We might eventually also want to
+			// consider "DelegatedCode".
+			if (as<GeneratedCode>(f->code)) {
+				oldCode = f->codeRef->address();
+			}
+		}
+
+		// Detach the code refs.
+		if (f->code)
+			f->code->detach();
+		if (f->lookup)
+			f->lookup->detach();
+
+		// Create our references if needed.
+		if (f->codeRef || f->lookupRef)
+			initRefs();
+
+		// Steal the references.
+		if (f->codeRef)
+			codeRef->steal(f->codeRef);
+		if (f->lookupRef)
+			lookupRef->steal(f->lookupRef);
+
+		// Steal 'toReplace' so that we forward the backlog if any.
+		toReplace = f->toReplace;
+
+		f->codeRef = null;
+		f->lookupRef = null;
+		f->toReplace = null;
+
+		// See if we need to update our old version:
+		if (oldCode) {
+			addOldFunction(engine(), toReplace, oldCode);
+		}
+
+		// Anything old to replace, either from here or from old versions?
+		if (toReplace) {
+			tasks->replaceActive(this);
+		}
+	}
+
+	void Function::replaceActive(ReplaceTasks *tasks) {
+		// There are two failure modes here. One is if 'replaceActiveFunction' returns false.
+		// That is more of an implementation error, so we mostly ignore it. It may also throw an
+		// exception (e.g. if it fails to compile this function). If so, we want to remember the
+		// old function so that we can replace it in the future.
+
+		try {
+			// Note: we don't care if we fail to replace the function. That only means that the old
+			// version will run until completion, and the new one will be used for new calls.
+			if (toReplace) {
+				for (size_t i = 0; i < toReplace->count(); i++) {
+					const void *fn = toReplace->read(i);
+					if (fn)
+						replaceActiveFunction(fn, this, tasks);
+					toReplace->write(i, null);
+				}
+			}
+			// All of them are now updated, remove the array.
+			toReplace = null;
+
+		} catch (storm::Exception *e) {
+			// Store the error.
+			if (CodeError *c = as<CodeError>(e)) {
+				tasks->error(new (this) ActiveReplaceError(c, this));
+			} else {
+				tasks->error(e);
+			}
+		}
+	}
+
 	void Function::setCode(Code *code) {
 		if (this->code)
 			this->code->detach();
@@ -626,16 +747,70 @@ namespace storm {
 	}
 
 
+	MAYBE(Function *) Function::withDerivedThis(Type *derived) {
+		// Sanity checking:
+		if (params->empty())
+			return null;
+		Value first = params->at(0);
+		if (!derived->isA(first.type))
+			return null;
+		if (first.isValue() && !first.ref)
+			return null;
+
+		Array<Value> *paramCopy = new (this) Array<Value>(*params);
+		paramCopy->at(0).type = derived;
+
+		// Create a near-identical copy:
+		Function *wrap = new (this) Function(pos, result, name, paramCopy);
+		wrap->myFlags = myFlags;
+		wrap->visibility = visibility;
+		wrap->documentation = documentation;
+
+		// Set refs:
+		wrap->setCode(new (this) DelegatedCode(directRef()));
+		wrap->setLookup(new (this) DelegatedCode(ref()));
+
+		// Copy flags. Also set 'allowReplace' so that it can be replaced transparently.
+		wrap->flags = flags | namedAllowReplace;
+
+		return wrap;
+	}
+
+	void Function::automaticReplace(Named *old) {
+		Function *f = as<Function>(old);
+		if (!f)
+			return;
+
+		// Adopt the refs from the old function! We don't bother with active replacement here, since
+		// the user just added the new function to the name set.
+		if (f->code)
+			f->code->detach();
+		if (f->lookup)
+			f->lookup->detach();
+
+		if (f->codeRef || f->lookupRef)
+			initRefs();
+
+		if (f->codeRef)
+			codeRef->steal(f->codeRef);
+		if (f->lookupRef)
+			lookupRef->steal(f->lookupRef);
+
+		f->codeRef = null;
+		f->lookupRef = null;
+	}
+
+
 	/**
 	 * Low-level functions called by the generated code.
 	 */
 
 	void spawnThreadResult(const void *fn, bool member, os::CallThunk thunk, void **params, void *result, Thread *on) {
 		os::FnCallRaw call(params, thunk);
-		os::FutureSema<os::Sema> future;
+		os::FutureBase future;
 		const os::Thread *thread = on ? &on->thread() : null;
 		os::UThread::spawnRaw(fn, member, null, call, future, result, thread);
-		future.result();
+		future.result(&updateFutureExceptions, null);
 	}
 
 	void spawnThreadFuture(const void *fn, bool member, os::CallThunk thunk, void **params, FutureBase *result, Thread *on) {
