@@ -6,6 +6,7 @@
 #include "Adapter.h"
 #include "Exception.h"
 #include "Package.h"
+#include "TypeTransform.h"
 #include "Core/Str.h"
 #include "Core/Handle.h"
 #include "Core/Gen/CppTypes.h"
@@ -94,6 +95,19 @@ namespace storm {
 		init(null);
 	}
 
+	Type::Type(SrcPos pos, Str *name, Array<Value> *params, TypeFlags flags, Size size) :
+		NameSet(pos, name, params),
+		engine(RootObject::engine()),
+		myGcType(null),
+		tHandle(null),
+		myTypeFlags(flags | typeCpp),
+		mySize(size) {
+
+		if ((flags & typeValue) == 0)
+			throw new (this) InternalError(S("Can not use the Type constructor taking a Size to create class types."));
+		init(null);
+	}
+
 	Type::Type(Str *name, TypeFlags flags, Size size, GcType *gcType, const void *vtable) :
 		NameSet(name),
 		engine(RootObject::engine()),
@@ -161,14 +175,18 @@ namespace storm {
 		if (value())
 			return;
 
-		myVTable = new (engine) VTable(this);
+		VTable *vt = new (engine) VTable(this);
 		if (myTypeFlags & typeCpp) {
-			myVTable->createCpp(vtab);
+			vt->createCpp(vtab);
 		} else if (Type *t = super()) {
-			myVTable->createStorm(t->myVTable);
+			vt->createStorm(t->myVTable);
 		} else {
-			myVTable->createStorm(Object::stormType(engine)->myVTable);
+			vt->createStorm(Object::stormType(engine)->myVTable);
 		}
+
+		// Write it using atomics. That way we get sane behavior in relation to 'hasVTableOf' which
+		// may be called from any point in the system.
+		atomicWrite(myVTable, vt);
 	}
 
 	void Type::lateInit() {
@@ -275,10 +293,14 @@ namespace storm {
 			return Object::stormType(engine);
 	}
 
-	void Type::setSuper(MAYBE(Type *) to) {
+	void Type::setSuper(Type *to) {
+		setSuper(to, null);
+	}
+
+	void Type::setSuper(Type *to, ReplaceTasks *tasks) {
 		if (to && (to->typeFlags() & typeFinal)) {
 			StrBuf *msg = new (engine) StrBuf();
-			*msg << S("Can not inherit from the final type ") << to->identifier() << S(".");
+			*msg << S("Can not inherit the type ") << to->identifier() << S(" since it is marked 'final'.");
 			throw new (engine) TypedefError(pos, msg->toS());
 		}
 
@@ -316,11 +338,11 @@ namespace storm {
 		// topmost class we're moving, as all other vtables will be deleted.
 		vtableDetachedSuper(prev);
 
-		// Propagate chainges to all children.
-		updateSuper();
+		// Propagate changes to all children.
+		updateSuper(tasks);
 	}
 
-	void Type::updateSuper() {
+	void Type::updateSuper(ReplaceTasks *tasks) {
 		Type *to = super();
 
 		// Which thread to use?
@@ -343,6 +365,10 @@ namespace storm {
 			GcType *old = myGcType;
 			myGcType = null;
 			engine.gc.freeType(old);
+
+			// Create a new one if we are to generate 'replace' instructions.
+			if (tasks)
+				tasks->replace(old, gcType());
 		}
 
 		// TODO: Invalidate the Layout as well.
@@ -359,7 +385,7 @@ namespace storm {
 		// Recurse into children.
 		TypeChain::Iter i = chain->children();
 		while (Type *c = i.next())
-			c->updateSuper();
+			c->updateSuper(tasks);
 	}
 
 	MAYBE(Type *) Type::declaredSuper() const {
@@ -466,11 +492,10 @@ namespace storm {
 		if (Function *f = as<Function>(item)) {
 			if (*f->name == CTOR)
 				updateCtor(f);
-			else if (*f->name != DTOR)
-				vtableFnAdded(f);
-
-			if (*f->name == DTOR)
+			else if (*f->name == DTOR)
 				updateDtor(f);
+			else
+				vtableFnAdded(f);
 
 			if (tHandle)
 				updateHandle(f);
@@ -484,6 +509,102 @@ namespace storm {
 		}
 	}
 
+	Bool Type::remove(Named *item) {
+		if (!NameSet::remove(item))
+			return false;
+
+		if (Function *f = as<Function>(item)) {
+			if (*f->name == CTOR)
+				updateCtor(f);
+			else if (*f->name == DTOR)
+				updateDtor(null);
+			else
+				vtableFnRemoved(f);
+		}
+
+		if ((typeFlags() & typeCpp) != typeCpp) {
+			// TODO: Update layout?
+		}
+
+		return true;
+	}
+
+	Named *Type::createBaseWrapper(Named *fromBase, SimplePart *query, Scope scope) {
+		// The problem we are trying to solve here is the following:
+		// If we are the first class that defines that we are tied to a specific thread,
+		// and we return a function from the base class (e.g. 'toS(StrBuf)'). Then, when
+		// calling the returned function, the call mechanisms will see that the thread to
+		// use is dynamic (since TObject is thread: any), and therefore insert an unexpected
+		// thread switch. To solve this, we detect these cases and transparently insert a
+		// wrapper. As of writing, this only happens for TObject -> first class, so it
+		// only happens for the functions in TObject.
+
+		// No point in doing anything if we don't have a superclass. We should never get caught here
+		// due to how the function is called, but it is nice to be a bit robust.
+		Type *super = this->super();
+		if (!super)
+			return fromBase;
+
+		// First: check if we need to bother doing anything at all.
+		// This logic is the same as in 'runOn', but with checks in a slightly different
+		// order. This is to avoid calling 'isA' if we know that no thread is specified
+		// here. Since we are only interested in the case where a thread is specified,
+		// this produces the same result as calling 'runOn'.
+		if (!useThread)
+			return fromBase;
+		if (!isA(TObject::stormType(engine)))
+			return fromBase;
+
+		// Similarly, we can check our parent class here directly. We know that both we and our
+		// parent is at least a TObject (assuming we *have* a parent). If they have the same idea of
+		// which thread to use, then we don't have to do anything. Note that this should be the case
+		// if both 'useThread' are non-null.
+		if (super->useThread == useThread)
+			return fromBase;
+
+		// We are only interested in non-static functions.
+		Function *baseFunction = as<Function>(fromBase);
+		if (!baseFunction)
+			return fromBase;
+		if (!baseFunction->isMember())
+			return fromBase;
+
+		// Also ignore ctors and dtors:
+		if (*baseFunction->name == CTOR || *baseFunction->name == DTOR)
+			return fromBase;
+
+		// Note: Some of these paths return 'null' instead of 'fromBase'. These are cases where we
+		// have determined that the lookup is likely looking for this particular function from
+		// another subclass. As such, we return 'null' to nudge the name lookup logic to find the
+		// proper function from the appropriate subclass instead.
+
+		// See if it makes sense to create a wrapper by looking at the first parameter:
+		if (query->params->empty())
+			return fromBase;
+		Type *t = query->params->at(0).type;
+		// Note: Letting cases where a parameter is Value() through is deliberate. It lets
+		// implementations use Value() as a placeholder in case they do something strange,
+		// and would like us to call 'matches' instead.
+		if (t && !t->isA(this))
+			return null; // See above
+
+		// At this point, we can create the wrapper, add it, and return it.
+		Function *wrapper = baseFunction->withDerivedThis(this);
+		if (!wrapper) {
+			WARNING(L"withDerivedThis failed!");
+			return fromBase;
+		}
+
+		// Double check if we did not pass the checks previously:
+		wrapper->parentLookup(this);
+		if (query->matches(wrapper, scope) >= 0) {
+			add(wrapper);
+			return wrapper;
+		} else {
+			return null; // See above
+		}
+	}
+
 	Named *Type::find(SimplePart *part, Scope source) {
 		if (Named *n = NameSet::find(part, source))
 			return n;
@@ -493,7 +614,7 @@ namespace storm {
 			return null;
 
 		if (Type *s = super())
-			return s->find(part, source);
+			return createBaseWrapper(s->find(part, source), part, source);
 		else
 			return null;
 	}
@@ -528,10 +649,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Switch threads...
 			Type *me = this;
-			os::Future<code::TypeDesc *, Semaphore> f;
+			os::Future<code::TypeDesc *> f(os::FutureMode::exclusive);
 			os::FnCall<code::TypeDesc *, 1> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::typeDesc), true, p, f, &t);
-			return f.result();
+			return f.result(&updateFutureExceptions, null);
 		}
 
 		// Double-check so that the desc is not created.
@@ -617,7 +738,7 @@ namespace storm {
 		if (Type *parent = super()) {
 			TypeDesc *desc = parent->typeDesc();
 			if (SimpleDesc *s = as<SimpleDesc>(desc))
-				merge(Offset(), pos, into, s);
+				storm::merge(Offset(), pos, into, s);
 			else
 				throw new (this) TypedefError(this->pos,
 											S("Can not produce a SimpleDesc when the parent type is not a simple type!"));
@@ -626,7 +747,7 @@ namespace storm {
 		Array<MemberVar *> *vars = variables();
 		for (Nat i = 0; i < vars->count(); i++) {
 			MemberVar *v = vars->at(i);
-			Offset offset = v->offset();
+			Offset offset = v->rawOffset();
 			Value type = v->type;
 
 			if (type.isObject() || type.ref) {
@@ -643,7 +764,7 @@ namespace storm {
 					into->at(pos) = p->v.move(offset);
 				pos++;
 			} else if (SimpleDesc *s = as<SimpleDesc>(original)) {
-				merge(offset, pos, into, s);
+				storm::merge(offset, pos, into, s);
 			} else if (ComplexDesc *c = as<ComplexDesc>(original)) {
 				UNUSED(c);
 				throw new (this) TypedefError(this->pos,
@@ -713,10 +834,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Switch threads...
 			Type *me = this;
-			os::Future<Size, Semaphore> f;
+			os::Future<Size> f(os::FutureMode::exclusive);
 			os::FnCall<Size, 1> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::size), true, p, f, &t);
-			return f.result();
+			return f.result(&updateFutureExceptions, null);
 		}
 
 		mySize = createSize();
@@ -757,10 +878,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Switch threads...
 			Type *me = this;
-			os::Future<VTable *, Semaphore> f;
+			os::Future<VTable *> f(os::FutureMode::exclusive);
 			os::FnCall<VTable *, 1> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::vtable), true, p, f, &t);
-			return f.result();
+			return f.result(&updateFutureExceptions, null);
 		}
 
 		// In the proper thread. The only thing we need to do is to make sure that the VTable is
@@ -787,10 +908,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Switch threads...
 			Type *me = this;
-			os::Future<const Handle *, Semaphore> f;
+			os::Future<const Handle *> f(os::FutureMode::exclusive);
 			os::FnCall<const Handle *, 1> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::handle), true, p, f, &t);
-			return *f.result();
+			return *f.result(&updateFutureExceptions, null);
 		}
 
 		if (!tHandle)
@@ -894,10 +1015,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Wrong thread, switch!
 			Type *me = this;
-			os::Future<void *, Semaphore> f;
+			os::Future<void *> f(os::FutureMode::exclusive);
 			os::FnCall<void *, 1> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::rawCopyConstructor), true, p, f, &t);
-			return (CopyCtorFn)f.result();
+			return (CopyCtorFn)f.result(&updateFutureExceptions, null);
 		}
 
 		// Find the copy constructor...
@@ -910,6 +1031,291 @@ namespace storm {
 		return rawCtor;
 	}
 
+	void Type::findEquivalentTypes(Named *old, ReplaceContext *ctx) {
+		if (Type *o = as<Type>(old))
+			ctx->addEquivalence(o, this);
+	}
+
+	class TypeCheckDiff : public NameDiff {
+	public:
+		TypeCheckDiff(Type *type, ReplaceContext *ctx) : type(type), ctx(ctx) {}
+
+		Type *type;
+
+		ReplaceContext *ctx;
+
+		virtual void added(Named *item) {
+			if (MemberVar *var = as<MemberVar>(item)) {
+				// We currently only support modifying the layout of class types:
+				if (type->isValue())
+					throw new (item) ReplaceError(item->pos, S("Unable to add member variables to value types."));
+
+				// Check so that we either have an initializer, or that the type in the variable has
+				// a default ctor we can use.
+				if (!var->initializer() && (var->type.type == null || var->type.type->defaultCtor() == null)) {
+					StrBuf *message = new (item) StrBuf();
+					*message << S("Unable to add the variable ") << item->name << S(" to ") << type->identifier()
+							 << S(" since no initializer was specified, and since the type ")
+							 << var->type << S(" does not have a default constructor. Add an initializer and try again.");
+					throw new (item) ReplaceError(item->pos, message->toS());
+				}
+			}
+		}
+		virtual void added(Template *) { /* We don't care */ }
+
+		virtual void removed(Named *item) {
+			if (as<MemberVar>(item)) {
+				// At the time being, we don't support removing members. We need to check if they are used first.
+				throw new (item) ReplaceError(item->pos, S("Unable to remove member variables from types."));
+			}
+		}
+		virtual void removed(Template *) { /* We don't care */ }
+
+		virtual void changed(Named *old, Named *changed) {
+			if (Str *msg = changed->canReplace(old, ctx))
+				throw new (changed) ReplaceError(changed->pos, msg);
+
+		}
+	};
+
+	MAYBE(Str *) Type::canReplace(Named *old, ReplaceContext *ctx) {
+		Type *o = as<Type>(old);
+		if (!o)
+			return new (this) Str(S("Unable to replace a non-type with a type."));
+
+		// Note: Due to type equivalences, the remainder of the system more or less assumes that all
+		// types can be replaced without issues. Therefore, we throw errors if we fail for some
+		// reason so that the system does not attempt to simply remove and then add the type.
+
+		// We only need to check for conflicts if the old entity has actually been loaded. If it is
+		// not loaded, we can replace it without any issues whatsoever.
+		if (!o->allLoaded())
+			return null;
+
+		forceLoad();
+
+		if (value() != o->value())
+			throw new (this) ReplaceError(pos, S("Can not change classes into values or vice versa."));
+
+		if (super() && !o->super()) {
+			throw new (this) ReplaceError(pos, S("Can not introduce a super class."));
+		} else if (!super() && o->super()) {
+			throw new (this) ReplaceError(pos, S("Can not remove a super class."));
+		} else if (super()) {
+			if (!ctx->same(super(), o->super()))
+				throw new (this) ReplaceError(pos, S("Can not change the super class."));
+		}
+
+		TypeCheckDiff diff(this, ctx);
+		o->diff(this, diff, ctx);
+
+		return null;
+	}
+
+	class TypeReplaceDiff : public NameDiff {
+	public:
+		TypeReplaceDiff(ReplaceTasks *tasks, ReplaceContext *ctx)
+			: tasks(tasks), context(ctx), membersUpdated(false) {}
+
+		ReplaceTasks *tasks;
+		ReplaceContext *context;
+
+		Bool membersUpdated;
+
+		virtual void added(Named *item) {
+			if (as<MemberVar>(item))
+				membersUpdated = true;
+		}
+		virtual void added(Template *) { /* We don't care */ }
+
+		virtual void removed(Named *item) {
+			if (as<MemberVar>(item))
+				membersUpdated = true;
+		}
+		virtual void removed(Template *) { /* We don't care */ }
+
+		virtual void changed(Named *old, Named *changed) {
+			// Do the replacement.
+			changed->replace(old, tasks, context);
+
+			// If both are variables, see if the offset changed. If so, we need to update the heap.
+			if (MemberVar *oldVar = as<MemberVar>(old)) {
+				if (MemberVar *newVar = as<MemberVar>(changed)) {
+					if (oldVar->rawOffset() != newVar->rawOffset())
+						membersUpdated = true;
+				}
+			}
+		}
+	};
+
+	static Nat findVar(Array<MemberVar *> *in, MemberVar *item) {
+		for (Nat i = 0; i < in->count(); i++) {
+			if (*item->name == *in->at(i)->name)
+				return i;
+		}
+		return in->count();
+	}
+
+	static Bool hasUpdatedTransform(Type *type, ReplaceTasks *tasks, ReplaceContext *ctx, Bool checkMembers = false) {
+		Type *oldType = ctx->normalize(type);
+		// No updated transform if the type has not been replaced:
+		if (oldType == type)
+			return false;
+
+		// See if there is already an added transform for the type. This means that the parent's
+		// update ran before us, and it already added a transform for us.
+		if (tasks->hasTransformFor(oldType))
+			return true;
+
+		// If not, the parent might run after us.
+
+		// If 'checkMembers', we should check members here. This is to actually diff the members in
+		// the case the parent runs after a child.
+		if (checkMembers) {
+			// This is a simple check of members:
+			type->doLayout();
+			Array<MemberVar *> *oldVars = oldType->variables();
+			Array<MemberVar *> *newVars = type->variables();
+
+			if (oldVars->count() != newVars->count())
+				return true;
+			for (Nat oldI = 0; oldI < oldVars->count(); oldI++) {
+				MemberVar *oldVar = oldVars->at(oldI);
+				Nat newI = findVar(newVars, oldVar);
+				if (newI >= newVars->count())
+					return true;
+
+				MemberVar *newVar = newVars->at(newI);
+				if (oldVar->rawOffset() != newVar->rawOffset())
+					return true;
+			}
+		}
+
+		// If it was equal here, check the parent as well.
+		Type *super = type->super();
+		if (!super)
+			return false;
+
+		return hasUpdatedTransform(super, tasks, ctx, true);
+	}
+
+	void Type::doReplace(Named *old, ReplaceTasks *tasks, ReplaceContext *ctx) {
+		Type *o = (Type *)old;
+
+		// Force us to compute our layout now. We need it in the diff.
+		size(); // Calls 'doLayout' if it is necessary.
+
+		// Update derived classes (Note: We don't need to update our super class).
+		if (o->chain) {
+			TypeChain::Iter iter = o->chain->children();
+			while (Type *child = iter.next()) {
+				if (child->super() != this)
+					child->setSuper(this, tasks);
+			}
+		}
+
+		// TODO: We might be able to move some of this into NameSet.
+		TypeReplaceDiff diff(tasks, ctx);
+		o->diff(this, diff, ctx);
+
+		// Update the reference.
+		if (o->selfRef) {
+			typeRef(); // Make sure it is created in here.
+			selfRef->steal(o->selfRef);
+			o->selfRef = null;
+		}
+
+		// We need to replace references to the old one with references to us.
+		tasks->replace(old, this);
+
+		// Also replace handles.
+		if (o->tHandle) {
+			// For many reference types, the handles are the same.
+			if (o->tHandle != &handle())
+				tasks->replace(o->tHandle, tHandle);
+		}
+
+		// Here, we have two options depending on whether the layout changed or not:
+		if (diff.membersUpdated) {
+			// Expensive path. We need to rewrite all instances of this type on the heap.
+			createTypeTransform(tasks, ctx);
+		} else if (!hasUpdatedTransform(this, tasks, ctx)) {
+			// Less expensive path. The object did not change, but we need to update the GcType of
+			// all objects so that the Type returned is correct, and update the VTables.
+			if (o->myGcType)
+				tasks->replace(o->myGcType, gcType());
+			if (o->myVTable)
+				vtable()->replace(o->myVTable, tasks);
+		}
+	}
+
+	void Type::createTypeTransformRec(ReplaceTasks *tasks, ReplaceContext *ctx) {
+		createTypeTransform(tasks, ctx);
+
+		if (!chain)
+			return;
+
+		TypeChain::Iter children = chain->children();
+		while (Type *t = children.next())
+			t->createTypeTransformRec(tasks, ctx);
+	}
+
+	void Type::createTypeTransform(ReplaceTasks *tasks, ReplaceContext *ctx) {
+		Type *oldType = ctx->normalize(this);
+		if (tasks->hasTransformFor(oldType))
+			return;
+
+		TypeTransform *tfm = new (this) TypeTransform(oldType, this);
+		populateTypeTransform(tfm, ctx);
+		tasks->transform(tfm);
+	}
+
+	void Type::populateTypeTransform(TypeTransform *tfm, ReplaceContext *ctx) {
+		if (Type *s = super())
+			s->populateTypeTransform(tfm, ctx);
+
+		Type *oldType = ctx->normalize(this);
+
+		// If this type has not been replaced, we don't have to do anything more.
+		if (oldType == this)
+			return;
+
+		// No layout for the old type -> declared in C++.
+		if (!oldType->layout)
+			return;
+
+		// Force the layout to happen now.
+		size();
+		assert(layout);
+
+		// Find differences between variables here:
+		Array<MemberVar *> *oldVars = oldType->layout->variables();
+		Array<MemberVar *> *newVars = layout->variables();
+
+		for (Nat oldId = 0; oldId < oldVars->count(); oldId++) {
+			MemberVar *oldVar = oldVars->at(oldId);
+
+			Nat newId = findVar(newVars, oldVar);
+			if (newId >= newVars->count()) {
+				// Not found, register it as a removed variable.
+				tfm->removed(oldVar);
+			} else {
+				// Found, register it as a changed variable.
+				tfm->same(oldVar, newVars->at(newId));
+			}
+		}
+
+		for (Nat newId = 0; newId < newVars->count(); newId++) {
+			MemberVar *newVar = newVars->at(newId);
+
+			Nat oldId = findVar(oldVars, newVar);
+			if (oldId >= oldVars->count()) {
+				// Not found, it is a new variable.
+				tfm->added(newVar);
+			}
+		}
+	}
+
 	void Type::useSuperGcType() {
 		if (myGcType)
 			return;
@@ -972,10 +1378,10 @@ namespace storm {
 		if (t != os::Thread::current()) {
 			// Wrong thread, switch!
 			Type *me = this;
-			os::Future<const GcType *, Semaphore> f;
+			os::Future<const GcType *> f(os::FutureMode::exclusive);
 			os::FnCall<const GcType *, 2> p = os::fnCall().add(me);
 			os::UThread::spawn(address(&Type::gcType), true, p, f, &t);
-			return f.result();
+			return f.result(&updateFutureExceptions, null);
 		}
 
 		// We're on the correct thread. Compute the type!
@@ -1166,8 +1572,14 @@ namespace storm {
 	}
 
 	void Type::updateHandle(Function *fn) {
-		bool val = value();
+		if (!as<const RefHandle>(tHandle)) {
+			// This happens if we are a TObject. Then we get a default handle. Without this check,
+			// we would crash when a 'hash' function is defined, for example.
+			return;
+		}
+
 		RefHandle *h = (RefHandle *)tHandle;
+		bool val = value();
 		Array<Value> *params = fn->params;
 		Str *name = fn->name;
 
@@ -1312,7 +1724,7 @@ namespace storm {
 		*src << code::fnRetRef(param);
 
 		readRef = dynamicFunction(engine, r, S("_read_"), valList(engine, 1, r.asRef()), src);
-		readRef->parentLookup = engine.package();
+		readRef->parentLookup(engine.package());
 		return readRef;
 	}
 
@@ -1384,6 +1796,22 @@ namespace storm {
 			myVTable->insert(fn);
 	}
 
+	void Type::vtableFnRemoved(Function *fn) {
+		// If we are a value, we do not have a vtable.
+		if (value())
+			return;
+
+		// Don't care if it is a static function.
+		if (!fn->isMember())
+			return;
+
+		// See if we can remove the function from the vtable.
+		if (myVTable)
+			myVTable->remove(fn);
+
+		// TODO: Check if any 'abstract' and 'override' constraints were broken.
+	}
+
 	// See if the return type of a function matches the super function.
 	// If the result differs "too much" in some sense, the calling conventions of the two functions
 	// will differ, and we will either produce weird results or crash.
@@ -1488,7 +1916,8 @@ namespace storm {
 		if (!engine.has(bootTemplates))
 			return;
 
-		old->myVTable->removeChild(this);
+		if (old->myVTable)
+			old->myVTable->removeChild(this);
 		isAbstract = abstractUnknown;
 	}
 
