@@ -16,35 +16,42 @@ namespace storm {
 		return nat(ms);
 	}
 
+	static void attach(os::Handle h, os::Thread &attached) {
+		if (attached == os::Thread::invalid) {
+			attached = os::Thread::current();
+			attached.attach(h);
+		}
+	}
+
 	/**
 	 * System-specific helpers. These all behave as if the handle was blocking.
 	 */
 
 #if defined(WINDOWS)
 
-	static inline void close(os::Handle &h, os::Thread &attached) {
-		CloseHandle(h.v());
-		attached = os::Thread::invalid;
+	void close(os::Handle &h, os::Thread &attached) {
+		os::Handle toClose = h;
 		h = os::Handle();
+
+		CloseHandle(toClose.v());
 	}
 
 	static PeekReadResult read(os::Handle h, os::Thread &attached, void *dest, Nat limit,
-							sys::ErrorCode &errorCode, Duration timeout = Duration()) {
-		if (attached == os::Thread::invalid) {
-			attached = os::Thread::current();
-			attached.attach(h);
-		}
+							Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
+		attach(h, attached);
 
 		os::IORequest request(h, attached, toMs(timeout));
 
 		LARGE_INTEGER pos;
 		pos.QuadPart = 0;
-		if (SetFilePointerEx(h.v(), pos, &pos, FILE_CURRENT)) {
+		if (!noPos && SetFilePointerEx(h.v(), pos, &pos, FILE_CURRENT)) {
 			// There seems to be a poblem when reading from the end of a file asynchronously.
 			LARGE_INTEGER len;
+			len.QuadPart = 0;
 			GetFileSizeEx(h.v(), &len);
-			if (pos.QuadPart >= len.QuadPart)
+			if (pos.QuadPart >= len.QuadPart) {
 				return PeekReadResult::end();
+			}
 
 			request.Offset = pos.LowPart;
 			request.OffsetHigh = pos.HighPart;
@@ -65,9 +72,11 @@ namespace storm {
 			// Completing async...
 			request.wake.down();
 
-			// Advance the file pointer.
-			pos.QuadPart = request.bytes;
-			SetFilePointerEx(h.v(), pos, NULL, FILE_CURRENT);
+			if (!noPos) {
+				// Advance the file pointer.
+				pos.QuadPart = request.bytes;
+				SetFilePointerEx(h.v(), pos, NULL, FILE_CURRENT);
+			}
 
 			// Was the operation cancelled?
 			if (request.bytes == 0 && request.error == ERROR_OPERATION_ABORTED)
@@ -85,17 +94,15 @@ namespace storm {
 		return PeekReadResult::success(request.bytes);
 	}
 
-	static Nat write(os::Handle h, os::Thread &attached, const void *src, Nat limit, sys::ErrorCode &errorCode) {
-		if (attached == os::Thread::invalid) {
-			attached = os::Thread::current();
-			attached.attach(h);
-		}
+	static Nat write(os::Handle h, os::Thread &attached, const void *src, Nat limit,
+					Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
+		attach(h, attached);
 
-		os::IORequest request(h, attached, 0);
+		os::IORequest request(h, attached, toMs(timeout));
 
 		LARGE_INTEGER pos;
 		pos.QuadPart = 0;
-		if (SetFilePointerEx(h.v(), pos, &pos, FILE_CURRENT)) {
+		if (!noPos && SetFilePointerEx(h.v(), pos, &pos, FILE_CURRENT)) {
 			// All is well.
 			request.Offset = pos.LowPart;
 			request.OffsetHigh = pos.HighPart;
@@ -117,8 +124,10 @@ namespace storm {
 			request.wake.down();
 
 			// Advance the file pointer.
-			pos.QuadPart = request.bytes;
-			SetFilePointerEx(h.v(), pos, NULL, FILE_CURRENT);
+			if (!noPos) {
+				pos.QuadPart = request.bytes;
+				SetFilePointerEx(h.v(), pos, NULL, FILE_CURRENT);
+			}
 
 			// Error?
 			if (request.error != 0)
@@ -170,9 +179,100 @@ namespace storm {
 		return dupHandle(GetStdHandle(id));
 	}
 
+#elif defined(LINUX_IO_URING)
+
+	void close(os::Handle &h, os::Thread &attached) {
+		if (h != os::Handle()) {
+			if (attached != os::Thread::invalid) {
+				// Detach and post a close request as a single operation.
+				os::IORequest r(h, attached);
+				r.request.opcode = IORING_OP_CLOSE;
+
+				// submitAndDetach may block, so update 'h' already here so that we don't try to
+				// close the same fd multiple times.
+				h = os::Handle();
+				r.submitAndDetach();
+			}
+		}
+
+		attached = os::Thread::invalid;
+		h = os::Handle();
+	}
+
+	static PeekReadResult read(os::Handle h, os::Thread &attached, void *dest, Nat limit,
+							Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
+		attach(h, attached);
+
+		os::IORequest r(h, attached);
+
+		r.request.opcode = IORING_OP_READ;
+		r.request.addr = reinterpret_cast<size_t>(dest);
+		r.request.len = limit;
+		r.request.off = -1; // Advances the file pointer as usual.
+
+		int result = r.submit(toMs(timeout));
+		// Note: ECANCELED means that the timeout was reached, or that the file was closed during
+		// the operation (since we cancel it then). We don't report that as an error, since none of
+		// the other implementations do.
+		if (result < 0 && result != -ECANCELED)
+			errorCode = fromSystemError(-result);
+
+		if (result <= 0)
+			return r.timeout ? PeekReadResult::timeout() : PeekReadResult::end();
+		else
+			return PeekReadResult::success(Nat(result));
+	}
+
+	static Nat write(os::Handle h, os::Thread &attached, const void *src, Nat limit,
+					Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
+		attach(h, attached);
+
+		os::IORequest r(h, attached);
+
+		r.request.opcode = IORING_OP_WRITE;
+		r.request.addr = reinterpret_cast<size_t>(src);
+		r.request.len = limit;
+		r.request.off = -1; // Advances the file pointer as usual.
+
+		int result = r.submit(toMs(timeout));
+		if (result < 0 && result != -ECANCELED)
+			errorCode = fromSystemError(-result);
+
+		if (result <= 0)
+			return 0;
+		else
+			return Nat(result);
+	}
+
+	static void seek(os::Handle h, Word to) {
+		lseek64(h.v(), to, SEEK_SET);
+	}
+
+	static Word tell(os::Handle h) {
+		off64_t r = lseek64(h.v(), 0, SEEK_CUR);
+		if (r < 0)
+			return 0;
+		return Word(r);
+	}
+
+	static Word length(os::Handle h) {
+		off64_t old = lseek64(h.v(), 0, SEEK_CUR);
+		if (old < 0)
+			return 0;
+
+		off64_t size = lseek64(h.v(), 0, SEEK_END);
+		lseek64(h.v(), old, SEEK_SET);
+		return Word(size);
+	}
+
+	static os::Handle dupHandle(os::Handle src) {
+		return os::Handle(dup(src.v()));
+	}
+
+
 #elif defined(POSIX)
 
-	static inline void close(os::Handle &h, os::Thread &attached) {
+	void close(os::Handle &h, os::Thread &attached) {
 		if (attached != os::Thread::invalid)
 			attached.detach(h);
 
@@ -188,10 +288,7 @@ namespace storm {
 
 	// Returns 'false' if the handle was closed (by us) during the operation.
 	static WaitResult doWait(os::Handle h, os::Thread &attached, os::IORequest::Type type, Duration timeout) {
-		if (attached == os::Thread::invalid) {
-			attached = os::Thread::current();
-			attached.attach(h);
-		}
+		attach(h, attached);
 
 		os::IORequest request(h, type, attached, toMs(timeout));
 		request.wake.wait();
@@ -203,7 +300,7 @@ namespace storm {
 	}
 
 	static PeekReadResult read(os::Handle h, os::Thread &attached, void *dest, Nat limit,
-							sys::ErrorCode &errorCode, Duration timeout = Duration()) {
+							Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
 		while (true) {
 			ssize_t r = ::read(h.v(), dest, size_t(limit));
 			if (r >= 0)
@@ -231,7 +328,7 @@ namespace storm {
 	}
 
 	static Nat write(os::Handle h, os::Thread &attached, const void *src, Nat limit,
-					sys::ErrorCode &errorCode) {
+					Bool noPos, sys::ErrorCode &errorCode, Duration timeout = Duration()) {
 		while (true) {
 			ssize_t r = ::write(h.v(), src, size_t(limit));
 			if (r >= 0)
@@ -243,7 +340,7 @@ namespace storm {
 				continue;
 			} else if (error == EAGAIN) {
 				// Wait for more data.
-				WaitResult r = doWait(h, attached, os::IORequest::write, Duration());
+				WaitResult r = doWait(h, attached, os::IORequest::write, timeout);
 				if (r.closed || r.timeout)
 					break;
 			} else {
@@ -292,11 +389,13 @@ namespace storm {
 	HandleIStream::HandleIStream(os::Handle h)
 		: handle(h),
 		  attachedTo(os::Thread::invalid),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	HandleIStream::HandleIStream(os::Handle h, os::Thread t)
 		: handle(h),
 		  attachedTo(t),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	// HandleIStream::HandleIStream(const HandleIStream &o)
@@ -305,7 +404,7 @@ namespace storm {
 	// 	  attachedTo(os::Thread::invalid) {}
 
 	HandleIStream::HandleIStream(const HandleIStream &o)
-		: PeekIStream(o), attachedTo(os::Thread::invalid), currError(o.currError) {
+		: PeekIStream(o), attachedTo(os::Thread::invalid), noClose(o.noClose), currError(o.currError) {
 		throw new (this) NotSupported(S("Copying HandleIStream"));
 	}
 
@@ -322,15 +421,16 @@ namespace storm {
 	}
 
 	void HandleIStream::close() {
-		if (handle)
+		if (handle && !noClose)
 			storm::close(handle, attachedTo);
+		handle = os::Handle();
 
 		PeekIStream::close();
 	}
 
 	PeekReadResult HandleIStream::doRead(byte *to, Nat count) {
 		if (handle)
-			return storm::read(handle, attachedTo, to, count, currError);
+			return storm::read(handle, attachedTo, to, count, noPos, currError);
 		else
 			return PeekReadResult::end();
 	}
@@ -342,18 +442,20 @@ namespace storm {
 	HandleRIStream::HandleRIStream(os::Handle h)
 		: handle(h),
 		  attachedTo(os::Thread::invalid),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	HandleRIStream::HandleRIStream(os::Handle h, os::Thread t)
 		: handle(h),
 		  attachedTo(t),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	// HandleRIStream::HandleRIStream(const HandleRIStream &o)
 	// 	: handle(dupHandle(o.handle)),
 	// 	  attachedTo(os::Thread::invalid) {}
 
-	HandleRIStream::HandleRIStream(const HandleRIStream &o) : attachedTo(os::Thread::invalid), currError(o.currError) {
+	HandleRIStream::HandleRIStream(const HandleRIStream &o) : attachedTo(os::Thread::invalid), noClose(false), currError(o.currError) {
 		throw new (this) NotSupported(S("Copying HandleIStream"));
 	}
 
@@ -382,7 +484,7 @@ namespace storm {
 		if (start >= b.count())
 			return b;
 
-		PeekReadResult r = storm::read(handle, attachedTo, b.dataPtr() + start, b.count() - start, currError);
+		PeekReadResult r = storm::read(handle, attachedTo, b.dataPtr() + start, b.count() - start, noPos, currError);
 		b.filled(r.bytesRead() + start);
 		return b;
 	}
@@ -399,8 +501,9 @@ namespace storm {
 	}
 
 	void HandleRIStream::close() {
-		if (handle)
+		if (handle && !noClose)
 			storm::close(handle, attachedTo);
+		handle = os::Handle();
 	}
 
 	RIStream *HandleRIStream::randomAccess() {
@@ -435,7 +538,7 @@ namespace storm {
 
 	PeekReadResult HandleTimeoutIStream::doRead(byte *to, Nat count) {
 		if (handle)
-			return storm::read(handle, attachedTo, to, count, currError, timeout);
+			return storm::read(handle, attachedTo, to, count, noPos, currError, timeout);
 		else
 			return PeekReadResult::end();
 	}
@@ -447,18 +550,21 @@ namespace storm {
 	HandleOStream::HandleOStream(os::Handle h)
 		: handle(h),
 		  attachedTo(os::Thread::invalid),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	HandleOStream::HandleOStream(os::Handle h, os::Thread t)
 		: handle(h),
 		  attachedTo(t),
+		  noClose(false),
 		  currError(sys::none) {}
 
 	// HandleOStream::HandleOStream(const HandleOStream &o)
 	// 	: handle(dupHandle(o.handle)),
 	// 	  attachedTo(os::Thread::invalid) {}
 
-	HandleOStream::HandleOStream(const HandleOStream &o) : attachedTo(os::Thread::invalid), currError(o.currError) {
+	HandleOStream::HandleOStream(const HandleOStream &o)
+		: attachedTo(os::Thread::invalid), noClose(o.noClose), currError(o.currError) {
 		throw new (this) NotSupported(S("Copying HandleIStream"));
 	}
 
@@ -472,7 +578,7 @@ namespace storm {
 		start = min(start, to.filled());
 		if (handle) {
 			while (start < to.filled()) {
-				Nat r = storm::write(handle, attachedTo, to.dataPtr() + start, to.filled() - start, currError);
+				Nat r = storm::write(handle, attachedTo, to.dataPtr() + start, to.filled() - start, noPos, currError);
 				if (r == 0)
 					break;
 				start += r;
@@ -483,8 +589,33 @@ namespace storm {
 	}
 
 	void HandleOStream::close() {
-		if (handle)
+		if (handle && !noClose)
 			storm::close(handle, attachedTo);
+		handle = os::Handle();
+	}
+
+	/**
+	 * With timeout.
+	 */
+
+	HandleTimeoutOStream::HandleTimeoutOStream(os::Handle h) : HandleOStream(h), timeout() {}
+
+	HandleTimeoutOStream::HandleTimeoutOStream(os::Handle h, os::Thread attachedTo)
+		: HandleOStream(h, attachedTo), timeout() {}
+
+	Nat HandleTimeoutOStream::write(Buffer to, Nat start) {
+		Nat consumed = 0;
+		start = min(start, to.filled());
+		if (handle) {
+			while (start < to.filled()) {
+				Nat r = storm::write(handle, attachedTo, to.dataPtr() + start, to.filled() - start, noPos, currError, timeout);
+				if (r == 0)
+					break;
+				start += r;
+				consumed += r;
+			}
+		}
+		return consumed;
 	}
 
 }
