@@ -11,6 +11,14 @@ namespace storm {
 	 */
 
 	static void checkCtor(Type *t, FnBase *ctor) {
+		if (!runtime::isValue(t) && t != StormInfo<Str>::type(ctor->engine())) {
+			// We assume that the ctor is an actual ctor, otherwise we can not handle cycles
+			// properly (= we need to create the allocation and put it in a hash table before
+			// calling the ctor). Ask the FnBase about the situation!
+			if (!ctor->rawCtor()) {
+				throw new (ctor) InternalError(S("Function pointers passed to the serialization system must refer to constructors."));
+			}
+		}
 		// TODO: Check!
 	}
 
@@ -165,6 +173,10 @@ namespace storm {
 	ObjIStream::Desc::Desc(Byte flags, Nat parent, Str *name) : data(Nat(flags) << 24), parent(parent) {
 		members = new (this) Array<Member>();
 
+		// If we want backwards compatibility with the old HttpProtocol we can do something like this:
+		// if (*name == S("core\x01io\x01HttpProtocol\x01"))
+		// 	name = new (name) Str(S("http\x01HttpProtocol\x01"));
+
 		Type *t = runtime::fromIdentifier(name);
 		if (!t)
 			throw new (this) SerializationFormatError(TO_S(this, S("Unknown type: ") << demangleName(name)));
@@ -285,13 +297,14 @@ namespace storm {
 	 */
 
 	template <class T, T (IStream::* readFn)()>
-	static void CODECALL read(T *out, ObjIStream *from) {
-		new (out) T((from->from->*readFn)());
+	static T CODECALL read(ObjIStream *from) {
+		T tmp = (from->from->*readFn)();
 		from->end();
+		return tmp;
 	}
 
-	static void readStr(Str **out, ObjIStream *from) {
-		new (Place(out)) Str(from);
+	static Str *CODECALL readStr(ObjIStream *from) {
+		return new (from) Str(from);
 	}
 
 	ObjIStream::ObjIStream(IStream *src)
@@ -311,7 +324,7 @@ namespace storm {
 		typeIds->put(id, new (this) Desc(								\
 						typeInfo::none,									\
 						StormInfo<t>::type(e),							\
-						new (e) FnBase(address(&read<t, &IStream::read ## t>), null, false, null)))
+						fnPtr(e, &read<t, &IStream::read ## t>, null)));
 
 		// Add built-in types here to avoid special cases later on.
 		Engine &e = engine();
@@ -327,8 +340,10 @@ namespace storm {
 #undef ADD_BUILTIN
 
 		// String is a special case.
-		typeIds->put(strId, new (this) Desc(typeInfo::classType, StormInfo<Str>::type(e),
-														new (e) FnBase(address(&readStr), null, false, null)));
+		typeIds->put(strId, new (this) Desc(
+						typeInfo::classType,
+						StormInfo<Str>::type(e),
+						fnPtr(e, &readStr, null)));
 	}
 
 	void ObjIStream::readValue(Type *type, PTR_GC out) {
@@ -354,7 +369,7 @@ namespace storm {
 			throw new (this) SerializationFormatError(msg);
 		}
 
-		readValueI(expected, out);
+		readValueI(expected, out, false);
 	}
 
 	Object *ObjIStream::readClass(Type *type) {
@@ -384,8 +399,8 @@ namespace storm {
 		if (expected->isValue()) {
 			// If this type has turned into a class type, just go ahead. There is no problem with
 			// this as long as the members match.
-			Object *created = (Object *)runtime::allocObject(0, expected->info->type);
-			readValueI(expected, created);
+			Object *created = null;
+			readValueI(expected, &created, true);
 			return created;
 		}
 
@@ -417,7 +432,7 @@ namespace storm {
 			return;
 		}
 
-		readValueI(expected, out);
+		readValueI(expected, out, false);
 	}
 
 	Object *ObjIStream::readPrimitiveObject(StoredId id) {
@@ -435,8 +450,8 @@ namespace storm {
 		if (id != info.expectedType)
 			throw new (this) SerializationFormatError(S("Mismatch of built-in types!"));
 
-		Object *created = (Object *)runtime::allocObject(0, expected->info->type);
-		readValueI(expected, created);
+		Object *created = null;
+		readValueI(expected, &created, true);
 		return created;
 	}
 
@@ -446,22 +461,22 @@ namespace storm {
 			Type *t = type->info->type;
 			if (runtime::isValue(t)) {
 				Variant v = Variant::uninitializedValue(t);
-				readValueI(type, v.getValue());
+				readValueI(type, v.getValue(), false);
 				v.valueInitialized();
 				return v;
 			} else {
 				// We support turning values into classes.
-				Object *created = (Object *)runtime::allocObject(0, t);
-				readValueI(type, created);
+				Object *created = null;
+				readValueI(type, &created, true);
 				return Variant(created);
 			}
 		} else {
-			// Note: Primitives are never duplicated, so we need to check if it is a user-type or a
-			// primitive type. This likely means that we don't de-duplicate strings. It is fine
+			// Note: Primitives are never de-duplicated, so we need to check if it is a user-type or
+			// a primitive type. This likely means that we don't de-duplicate strings. It is fine
 			// since strings are immutable, but it could be a bit wasteful.
 			if (typeId < firstCustomId) {
-				Object *created = (Object *)runtime::allocObject(0, type->info->type);
-				readValueI(type, created);
+				Object *created = null;
+				readValueI(type, &created, true);
 				return Variant(created);
 			} else {
 				return Variant(readClassI(type, type->info->type));
@@ -469,7 +484,7 @@ namespace storm {
 		}
 	}
 
-	void ObjIStream::readValueI(Desc *type, void *out) {
+	void ObjIStream::readValueI(Desc *type, void *out, Bool objectPtr) {
 		// Find the parent classes and push them on the stack to keep track of what we're doing.
 		Desc *d = type;
 		while (d != null) {
@@ -482,9 +497,23 @@ namespace storm {
 		}
 
 		// Call the constructor!
+		FnBase *readCtor = type->info->readCtor;
+
+		// Try to call the constructor directly if that is possible. This only works if the
+		// underlying function was already a constructor.
+		if (!objectPtr) {
+			typedef void (*CtorPtr)(void *, ObjIStream *);
+			CtorPtr ctor = (CtorPtr)readCtor->rawCtor();
+			if (ctor) {
+				(*ctor)(out, this);
+				return;
+			}
+		}
+
+		// Fallback, this always works:
 		ObjIStream *me = this;
-		os::FnCall<void, 2> call = os::fnCall().add(out).add(me);
-		type->info->readCtor->callRaw(call, null, null);
+		void *params[2] = { &me };
+		type->info->readCtor->rawCall().call(type->info->readCtor, out, params);
 	}
 
 	Object *ObjIStream::readClassI(Desc *expected, Type *t) {
@@ -508,11 +537,28 @@ namespace storm {
 		Type *type = actual->info->type;
 		checkAllocSize(Nat(runtime::typeGc(type)->stride));
 
-		// Allocate the object and deserialize by calling the constructor, just like we do with value types.
-		Object *created = (Object *)runtime::allocObject(0, type);
+		// Call the constructor. Here, we call the "raw" constructor to be able to keep track of cycles properly.
+
+		// Find the parent classes and push them on the stack to keep track of what we're doing.
+		{
+			// Note: Needs to be kept in sync with readValueI
+			Desc *d = actual;
+			while (d != null) {
+				depth->push(Cursor(d));
+
+				if (d->parent)
+					d = findInfo(d->parent);
+				else
+					d = null;
+			}
+		}
+
+		Object *created = (Object *)runtime::allocObject(0, actual->info->type);
 		objIds->put(objId, created);
 
-		readValueI(actual, created);
+		typedef void (*CtorFn)(void *, ObjIStream *);
+		CtorFn ctor = (CtorFn)actual->info->readCtor->rawCtor();
+		(*ctor)(created, this);
 
 		return created;
 	}
@@ -567,7 +613,7 @@ namespace storm {
 					init->rawCall().call(init, valueOut, params);
 				} else {
 					os::FnCall<Object *, 1> params = os::fnCall();
-					r.result = Variant(init->callRaw(params, null, null));
+					r.result = Variant(init->callRaw(params, 0, null));
 				}
 
 				return r;
