@@ -41,8 +41,12 @@ namespace storm {
 		for (Nat i = 0; i < frames->count(); i++) {
 			adapter.nextFrame();
 
-			Frame f = frames->at(i);
-			s.format(adapter, f.id, f.base, f.offset);
+			const Frame &f = frames->at(i);
+			if (f.id == Frame::nextThread()) {
+				*to << S("-- spawned from thread 0x") << hex(f.ptr()) << S(" --");
+			} else {
+				s.format(adapter, f.id, f.base, f.offset);
+			}
 		}
 	}
 
@@ -56,7 +60,11 @@ namespace storm {
 		for (Nat i = 0; i < trace.count(); i++) {
 			if (i > 0)
 				to << std::endl;
-			to << L"Frame " << std::setw(3) << i << L": at 0x" << trace[i].ptr();
+			to << std::setw(3) << i << L": ";
+			if (trace[i].id == StackTrace::Frame::nextThread())
+				to << L"-- spawned from thread 0x" << trace[i].ptr() << L" --";
+			else
+				to << L"at 0x" << trace[i].ptr();
 		}
 		return to;
 	}
@@ -65,7 +73,12 @@ namespace storm {
 		for (Nat i = 0; i < count(); i++) {
 			if (i > 0)
 				*to << S("\n");
-			*to << width(3) << i << S(": at 0x") << hex((*this)[i].ptr());
+			*to << width(3) << i << S(": ");
+			const Frame &frame = frames->at(i);
+			if (frame.id == Frame::nextThread())
+				*to << S("-- spawned from thread 0x") << hex(frame.ptr()) << S(" --");
+			else
+				*to << S("at 0x") << hex(frame.ptr());
 		}
 	}
 
@@ -107,4 +120,32 @@ namespace storm {
 		return result;
 	}
 
+	class StormAppendGen : public StormGen {
+	public:
+		Nat original;
+
+		StormAppendGen(StackTrace &to) : StormGen(to), original(to.count()) {}
+
+		void init(size_t count) {
+			if (count > 0)
+				result.reserve(Nat(original + count));
+		}
+
+		void put(const StackFrame &frame) {
+			// Overwrite the first stack frame.
+			if (result.count() == original) {
+				result.push(StackTrace::Frame(os::UThread::current().threadData(),
+														0,
+														StackTrace::Frame::nextThread()));
+			} else {
+				StormGen::put(frame);
+			}
+		}
+	};
+
+	void appendStackTrace(StackTrace &to, Nat ignore) {
+		StormAppendGen gen(to);
+		createStackTrace(gen, max(Nat(1), ignore) - 1, null);
+	}
+
 }
