@@ -9,9 +9,8 @@ namespace storm {
 	/**
 	 * A set of weak references for use in Storm and C++.
 	 *
-	 * This hash set only manages references to actor objects (ie. TObjects), as other objects are
-	 * usually not relevant to keep weak references to (as they are copied). This simplifies the
-	 * implementation of this set a great deal as well.
+	 * This hash set manages references to objects, and always hash based on pointer values since
+	 * weakness is a property of individual objects rather than a property of "values".
 	 *
 	 * The implementation is inspired from the hash map implementation found in Lua, see Map.h and
 	 * Set.h for details.
@@ -57,15 +56,15 @@ namespace storm {
 		 */
 
 		// Put a value. Returns 'true' if the value did not exist before.
-		Bool CODECALL putRaw(TObject *key);
+		Bool CODECALL putRaw(RootObject *key);
 
 		// Contains?
-		Bool CODECALL hasRaw(TObject *key);
+		Bool CODECALL hasRaw(RootObject *key);
 
-		// Note: we're missing 'get' and 'at', as they are useless for TObjects.
+		// Note: we're missing 'get' and 'at', as they are useless for pointer based hashes.
 
 		// Remove a value from the set.
-		Bool CODECALL removeRaw(TObject *key);
+		Bool CODECALL removeRaw(RootObject *key);
 
 		/**
 		 * Debug and benchmarking functions.
@@ -105,7 +104,7 @@ namespace storm {
 
 		// Allocated memory.
 		GcArray<Info> *info;
-		GcWeakArray<TObject> *data;
+		GcWeakArray<RootObject> *data;
 
 		// Watch out for moving objects.
 		GcWatch *watch;
@@ -123,24 +122,24 @@ namespace storm {
 		void rehash(nat size);
 
 		// Do a re-hash while looking for an element. Assumes 'watch' is non-null, and that some object have moved.
-		nat rehashFind(nat size, TObject *key);
+		nat rehashFind(nat size, RootObject *key);
 
 		// Do a re-hash while removing an element. Assumes 'watch' is non-null, and that some object have moved.
-		bool rehashRemove(nat size, TObject *key);
+		bool rehashRemove(nat size, RootObject *key);
 
 		// Insert a node, given its hash is known (eg. when re-hashing). Assumes no other node with
 		// the same key exists, and will therefore always insert the element. Returns the slot
 		// inserted into. 'watch' is a slot that needs to be updated whenever a slot is moved.
-		nat insert(TObject *key, nat hash, nat &watch);
+		nat insert(RootObject *key, nat hash, nat &watch);
 
 		// Remove an element, ignoring any moved objects. Returns 'true' if an object was removed.
-		bool remove(TObject *key);
+		bool remove(RootObject *key);
 
 		// Find the current location of 'key', given 'hash'. Returns 'Info::free' if none exists.
-		nat findSlot(TObject *key, nat hash);
+		nat findSlot(RootObject *key, nat hash);
 
 		// Helper for 'findSlot'. Does not work properly if objects have moved.
-		nat findSlotI(TObject *key, nat hash);
+		nat findSlotI(RootObject *key, nat hash);
 
 		// Compute the primary slot for a node, given its hash.
 		nat primarySlot(nat hash) const;
@@ -156,7 +155,7 @@ namespace storm {
 
 		// Helper for copying arrays.
 		GcArray<Info> *copyArray(const GcArray<Info> *src);
-		GcWeakArray<TObject> *copyArray(const GcWeakArray<TObject> *src);
+		GcWeakArray<RootObject> *copyArray(const GcWeakArray<RootObject> *src);
 
 	public:
 
@@ -182,11 +181,11 @@ namespace storm {
 			Iter(WeakSetBase *owner);
 
 			// Get the next element, or null if at the end.
-			MAYBE(TObject *) CODECALL nextRaw();
+			MAYBE(RootObject *) CODECALL nextRaw();
 
 		private:
 			// The gc array from the set.
-			GcWeakArray<TObject> *data;
+			GcWeakArray<RootObject> *data;
 
 			// Current position.
 			Nat pos;
@@ -204,8 +203,6 @@ namespace storm {
 
 	/**
 	 * C++ interface.
-	 *
-	 * TODO: Set vtables for class in constructors.
 	 */
 	template <class K>
 	class WeakSet : public WeakSetBase {
