@@ -7,6 +7,7 @@
 #include "Allocator.h"
 #include "Gc/License.h"
 #include "Gc/Root.h"
+#include "Gc/Walker.h"
 
 namespace storm {
 
@@ -91,8 +92,7 @@ namespace storm {
 		void endRamp();
 
 		// Walk the heap.
-		typedef void (*WalkCb)(RootObject *inspect, void *param);
-		void walkObjects(WalkCb fn, void *param);
+		void walk(Walker &context);
 
 		typedef GcRoot Root;
 
@@ -130,6 +130,9 @@ namespace storm {
 
 		// Throw an error as is appropriate for the current situation.
 		void throwError(const wchar *message);
+
+		// Walk the arena.
+		void walkI(smm::ArenaTicket &ticket, Walker &context);
 	};
 
 }
