@@ -136,9 +136,8 @@ namespace gui {
 	App *app(EnginePtr e) {
 		App *&v = appData(e.v);
 		if (!v) {
-			// Try to be threadsafe.
-			static util::Lock l;
-			util::Lock::L z(l);
+			// Thread safety.
+			os::Lock::L z(dataLock(e.v));
 			if (!v) {
 				// Poke the thread as well since we might be running on another OS thread.
 				Ui::thread(e.v)->thread();
