@@ -22,6 +22,7 @@
 (defvar storm-mode-stdlib nil "Root of the Storm standard library (currently not used).")
 (defvar storm-mode-compile-compiler nil "Try to compile the compiler before starting it.")
 (defvar storm-mode-max-edits 20 "Maximum number of entries saved in the edit history.")
+(defvar storm-mode-show-server-buffer nil "Show the language server's buffer by default. Either t (always), 'start (show on start), or 'run (when running a command).")
 
 ;; State for all buffers using storm-mode.
 (defvar global-storm-mode nil "Is 'global-storm-mode' enabled? Use the function 'global-storm-mode' to enable/disable.")
@@ -55,7 +56,7 @@
 	buffer
       (find-file-noselect full-file))))
 
-(setq storm-error-regexp "^ *\\(p?[0-9]+[>:] ?\\)?@\\([^(\n\t]+\\)(\\([0-9]+\\)\\(-[0-9]+\\)?): [A-Za-z ]+ error:")
+(setq storm-error-regexp "^ *\\(p?[0-9]+[>:] ?\\)?@\\([^(\n\t]+\\)(\\([0-9]+\\)\\(-[0-9]+\\)?): \\([A-Za-z ]+ error:\\|here$\\)")
 
 ;; compile.el was updated in emacs 27. In the new version, the old way of adding custom hooks no
 ;; longer work properly. Instead, it provides a documented way of adding hooks.
@@ -264,14 +265,18 @@
 (defvar storm-mode-map
   (let ((map (make-keymap)))
     ;; Shortcuts for debugging the internal representation in Storm.
-    (define-key map "\C-cd" 'storm-debug-tree)
-    (define-key map "\C-cc" 'storm-debug-content)
-    (define-key map "\C-cu" 'storm-debug-re-color)
-    (define-key map "\C-cr" 'storm-debug-re-open)
-    (define-key map "\C-cx" 'storm-debug-un-color)
-    (define-key map "\C-ce" 'storm-debug-find-error)
+    (define-key map "\C-cdc" 'storm-debug-content)
+    (define-key map "\C-cdt" 'storm-debug-tree)
+    (define-key map "\C-cdu" 'storm-debug-re-color)
+    (define-key map "\C-cdo" 'storm-debug-re-open)
+    (define-key map "\C-cdx" 'storm-debug-un-color)
+    (define-key map "\C-cde" 'storm-debug-find-error)
     (define-key map "}"     'storm-insert-indent)
     (define-key map "\C-ch" 'storm-doc)
+    (define-key map "\C-ce" 'storm-repl-eval)
+    (define-key map "\C-cr" 'storm-run-at-point)
+    (define-key map "\C-c\C-r" 'storm-run-all)
+    (define-key map "\C-cu" 'storm-reload-current)
     map)
   "Keymap for storm-mode")
 
@@ -338,6 +343,160 @@
   (when storm-buffer-id
     (storm-send (list 'error storm-buffer-id))))
 
+(defvar storm-available-repl nil "Available REPLs in the running Storm instance.")
+(defvar storm-eval-history nil "History for evals.")
+
+(defun storm-repl-list ()
+  "List all available REPLs for the currently running Storm process."
+  (interactive)
+  (when (eq storm-available-repl nil)
+    (let ((answer (storm-query '(repl-available) 5)))
+      (when (and (listp answer)
+		 (stringp (car answer)))
+	(setq storm-available-repl answer))))
+  (when (called-interactively-p)
+    (message "Available REPLs: %s" (cl-reduce (lambda (x y) (message "%S %S" x y) (concat x ", " y)) storm-available-repl)))
+  storm-available-repl)
+
+(defun storm-repl-eval (str &optional repl)
+  "Evaluate the supplied string in the default Storm repl, or 'repl' if specified."
+  (interactive
+   (let ((repl (storm-repl-list)))
+     (when (null repl)
+       (error "No repl supported by the current Storm process!"))
+     (list (read-string (concat (car repl) "> ") "" 'storm-eval-history))))
+
+  (when (not repl)
+    (setq repl (storm-repl-list))
+    (when (null repl)
+      (error "No repl supported by the current Storm process!"))
+    (setq repl (car repl)))
+
+  (storm-output-string (format "%s> %s\n" repl str))
+  (storm-send (list 'repl-eval repl str (buffer-file-name))))
+
+(defun storm-on-eval (params)
+  "Called when we received the result from an earlier eval command."
+  (let* ((ok-msg (car params))
+	 (err-msg (if (cdr params)
+		      (cadr params)
+		    nil))
+	 (short-msg (if ok-msg
+			(format "=> %s" ok-msg)
+		      (if err-msg
+			  (format "=> %s" err-msg)
+			nil)))
+	 (long-msg (if (and err-msg (string-prefix-p "@" err-msg))
+		       (format "=> Error:\n%s" err-msg)
+		     short-msg)))
+    (when short-msg
+      (message short-msg))
+    (when long-msg
+      (storm-output-string (concat long-msg "\n") (if ok-msg nil 'storm-msg-error))))
+  nil)
+
+(defun storm-run-at-point (point)
+  "Try to execute the function at the current point in the current buffer."
+  (interactive "d")
+  (let ((id (if storm-buffer-id
+		storm-buffer-id
+	      (buffer-file-name))))
+
+    (storm-send (list 'run id (1- point))))
+  (storm-show-server-buffer 'run))
+
+
+(defun storm-run-all ()
+  "Execute all functions in the current buffer."
+  (interactive)
+  (let ((id (if storm-buffer-id
+		storm-buffer-id
+	      (buffer-file-name))))
+
+    (storm-send (list 'run id)))
+  (storm-show-server-buffer 'run))
+
+(defun storm-on-run (params)
+  "Called when we received the result from an earlier run command."
+  (storm-on-eval params))
+
+(defun storm-reload-current ()
+  "Reload the current buffer. If it is in storm-mode, reload with the current contents, otherwise ask to save first."
+  (interactive)
+  (storm-reload-buffer (current-buffer)))
+
+(defun storm-same-directory (dir-a dir-b)
+  ;; Note: file-name-case-insensitive-p is not available until emacs 26.3. Basic string
+  ;; compare is good enough in most cases, we will just miss some "save" prompts.
+  (if (and (fboundp 'file-name-case-insensitive-p) (file-name-case-insensitive-p dir-a))
+      (string= (downcase dir-a) (downcase dir-b))
+    (string= dir-a dir-b)))
+
+(defun storm-reload-package ()
+  "Reload all files in the current package. This does not use the cached values in storm-mode buffers, so it will save all buffers."
+  (interactive)
+  (let ((path (file-name-directory (buffer-file-name))))
+
+    ;; Save relevant buffers.
+    (save-some-buffers nil
+		       (lambda ()
+			 (if (buffer-file-name)
+			     (storm-same-directory path
+						   (file-name-directory (buffer-file-name)))
+			   nil)))
+
+    ;; Clear compile.el cache:
+    (storm-clear-compilation-cache-dir path)
+
+    ;; Send a message.
+    (storm-send (list 'reload path))))
+
+
+(defun storm-reload-buffer (&rest buffers)
+  "Reload all buffers given as parameters. If a buffer is in storm-mode, reload it with the current contents, otherwise ask to save it first."
+  ;; Ask if any of the buffers in 'buffers' should be saved (only if they are not in storm-mode).
+  (save-some-buffers nil
+		     (lambda ()
+		       (if storm-buffer-id
+			   nil ;; Never needs to be saved if it is in storm-mode!
+			 (seq-contains buffers (current-buffer) #'eq)))) ;; Otherwise, only save if in 'buffers'
+
+  ;; Clear compilation cache from compile.el.
+  (storm-clear-compilation-cache (mapcar #'buffer-file-name buffers))
+
+  (let ((cmd (mapcar (lambda (buffer)
+		       (with-current-buffer buffer
+			   (if storm-buffer-id
+			       storm-buffer-id
+			     (buffer-file-name))))
+		     buffers)))
+    (storm-send (cons 'reload cmd))))
+
+(defun storm-clear-compilation-cache (files)
+  "Clear the compilation cache for all files in the parameter list.
+This is necessary when reloading files, otherwise errors from compile.el will not work properly."
+
+  (let ((compile-buffer (get-buffer "*compilation*")))
+    (when compile-buffer
+      (with-current-buffer compile-buffer
+	(mapc #'compilation--flush-file-structure files)))))
+
+(defun storm-clear-compilation-cache-dir (path)
+  "Clear compilation cache for all files in some directory."
+  (let ((compile-buffer (get-buffer "*compilation*"))
+	(to-update '()))
+    (when compile-buffer
+      (with-current-buffer compile-buffer
+	;; This is technically an internal variable, but should be fine.
+	(maphash (lambda (k v)
+		   (when (= 1 (length k))
+		     (when (storm-same-directory (file-name-directory (car k)) path)
+		       (setq to-update (cons (car k) to-update)))))
+		 compilation-locs)))
+
+    (when (not (null to-update))
+      (storm-clear-compilation-cache to-update))))
+
 ;; Convenience for highlighting.
 
 (defun storm-set-color (from to face)
@@ -545,7 +704,9 @@
 (defvar storm-process-wait nil "If symbol: waiting for a message with the given header.")
 (defvar storm-messages
   (let ((map (make-hash-table)))
-    (puthash 'color 'storm-on-color map)
+    (puthash 'color #'storm-on-color map)
+    (puthash 'repl-eval #'storm-on-eval map)
+    (puthash 'run #'storm-on-run map)
     map)
   "Messages handled by this plugin.")
 
@@ -553,6 +714,9 @@
   "Make sure that the background Storm process is up and running. Start it if neccessary."
   (interactive)
   (unless (storm-running-p)
+    ;; Make sure to terminate Storm gracefully along with Emacs.
+    (add-hook 'kill-emacs-hook #'storm-kill-emacs-hook)
+
     ;; TODO: Ask for missing parameters!
     (if storm-mode-compile-compiler
 	(storm-start-compile)
@@ -602,6 +766,17 @@
   (interactive)
   (storm-stop 'storm-start))
 
+(defun storm-kill-emacs-hook ()
+  "Hook called when emacs is terminating. Attempts to terminate Storm gracefully if it is running."
+  (when storm-process
+    (storm-send '(quit))
+    ;; Wait for the process to terminate, a maximum of about 1 second.
+    (let ((start-time (current-time))
+	  (timeout 1))
+      (while (and (process-live-p storm-process)
+		  (< (float-time (time-since start-time)) 1))
+	(sleep-for 0 10)))))
+
 (defun storm-send (message &rest force)
   "Send a message to the Storm process (launching it if it is not running)."
   (if (storm-running-p)
@@ -611,8 +786,10 @@
 		(cons message storm-process-message-queue))
 	(process-send-string storm-process (storm-encode message)))
     (progn
-      (setq storm-process-message-queue
-	    (cons message storm-process-message-queue))
+      ;; Don't cache "open" messages. We keep track of them anyway by re-opening all buffers later.
+      (unless (eq (car message) 'open)
+	(setq storm-process-message-queue
+	      (cons message storm-process-message-queue)))
       (storm-start))))
 
 (defun storm-query (message &optional timeout)
@@ -727,6 +904,7 @@
   (setq storm-process-id-to-sym (make-hash-table))
   (setq storm-process-next-id 0)
   (setq storm-mode-types (make-hash-table :test 'equal))
+  (setq storm-available-repl nil)
   (with-current-buffer storm-process-output
     (buffer-disable-undo)
     (setq buffer-read-only nil)
@@ -745,8 +923,17 @@
   (storm-send-messages storm-process-message-queue)
   (setq storm-process-message-queue nil)
 
+  ;; Show the buffer if we were asked to.
+  (storm-show-server-buffer 'start)
+
   (run-hooks 'storm-started-hook))
 
+(defun storm-show-server-buffer (event)
+  "Show the server buffer if we were asked to do so."
+  (when (or (eq storm-mode-show-server-buffer t) (eq storm-mode-show-server-buffer event))
+    (display-buffer storm-process-output '(nil (allow-no-window . t))))
+  nil)
+
 
 (defun storm-send-messages (queue)
   "Send all messages in the queue, in reverse order (as that is how they are added)."
