@@ -11,14 +11,27 @@
  * Fixed-size data entry for stack traces.
  */
 struct StackFrame {
-	StackFrame() : fnBase(null), offset(0), id(0) {}
+	StackFrame() : fnBase(null), offset(0), id(0), returnLocation(0) {}
 
 	// Base and offset for the function.
 	void *fnBase;
 	int offset;
 
+	// Get the function as a full pointer.
+	inline void *fn() const {
+		return (byte *)fnBase + offset;
+	}
+
 	// Which backend produced the data above.
 	int id;
+
+	// Location of the return adress on the stack for this frame. I.e. where is the address that
+	// *returns to* this frame stored in memory? Might be null.
+	void *returnLocation;
+
+	// Update the return location if it is non-null. On some architectures (e.g. ARM64) this
+	// function might need to sign the pointer as required by the ABI spec.
+	void updateReturnLocation(const void *target) const;
 };
 
 
