@@ -102,7 +102,7 @@ typedef struct AWLPoolStruct {
   Count succAccesses;       /* number of successive single accesses */
   FindDependentFunction findDependent; /*  to find a dependent object */
   awlStatTotalStruct stats;
-  Sig sig;                  /* <code/misc.h#sig> */
+  Sig sig;                  /* design.mps.sig.field.end.outer */
 } AWLPoolStruct, *AWL;
 
 
@@ -138,7 +138,7 @@ typedef struct AWLSegStruct {
   Count oldGrains;          /* grains allocated prior to last collection */
   Count singleAccesses;     /* number of accesses processed singly */
   awlStatSegStruct stats;
-  Sig sig;                  /* <code/misc.h#sig> */
+  Sig sig;                  /* design.mps.sig.field.end.outer */
 } AWLSegStruct, *AWLSeg;
 
 DECLARE_CLASS(Seg, AWLSeg, MutatorSeg);
@@ -1057,7 +1057,7 @@ static void awlSegReclaim(Seg seg, Trace trace)
   Bool hasBuffer = SegBuffer(&buffer, seg);
   Format format = pool->format;
   Count reclaimedGrains = (Count)0;
-  Count preservedInPlaceCount = (Count)0;
+  STATISTIC_DECL(Count preservedInPlaceCount = (Count)0)
   Size preservedInPlaceSize = (Size)0;
   Index i;
 
@@ -1089,7 +1089,7 @@ static void awlSegReclaim(Seg seg, Trace trace)
       AVER(BTGet(awlseg->scanned, i));
       BTSetRange(awlseg->mark, i, j);
       BTSetRange(awlseg->scanned, i, j);
-      ++preservedInPlaceCount;
+      STATISTIC(++preservedInPlaceCount);
       preservedInPlaceSize += AddrOffset(p, q);
     } else {
       BTResRange(awlseg->mark, i, j);
